#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on February 16, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QObject, Slot
from qtpy.QtWidgets import QPushButton

from karabo.common.api import walk_traits_object
from karabo.common.scenemodel.api import SceneModel, SceneTargetWindow
from karabogui import access as krb_access, icons
from karabogui.access import AccessRole, access_role_allowed, is_authenticated
from karabogui.controllers.util import load_extensions
from karabogui.dialogs.update_dialog import is_package_updated
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts)
from karabogui.indicators import get_user_session_button_data
from karabogui.logger import get_logger
from karabogui.mainwindow import CONFIGURATOR_TITLE, MainWindow, PanelAreaEnum
from karabogui.panels.api import (
    ConfigurationPanel, MacroPanel, ScenePanel, WidgetControllerPanel)
from karabogui.programs.utils import close_app
from karabogui.singletons.api import get_config, get_db_conn, get_project_model
from karabogui.util import (
    get_application_icon, move_to_cursor, process_qt_events, send_info)
from karabogui.wizards import TipsTricksWizard

GUI_EXTENSIONS = "GUIExtensions"


class PanelWrangler(QObject):
    """An object which handles wrangler duties for the myriad panels and
    windows shown in the GUI.

    Importantly, it also holds the reference to the main window.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.main_window = None
        self.splash = None
        self.connected_to_server = False
        self._editor = None
        # Panel containers
        # Project items (scenes, macros) {model: panel}
        self._project_item_panels = {}
        # Unattached scene panels {model: panel}
        self._unattached_scene_panels = {}

        # Unattached controller panels {`deviceId.path|Panel`: panel}
        self._widget_controller_panels = {}

        # Panels linked to instances {instance id: (panel, pos)}
        self._instance_panels = {}

        # Register to KaraboBroadcastEvent, Note: unregister_from_broadcasts is
        # not necessary due to the fact that the singleton mediator object and
        # `self` are being destroyed when the GUI exists
        event_map = {
            KaraboEvent.RestoreSceneView: self._event_restore_scene,
            KaraboEvent.ShowSceneView: self._event_attached_scene,
            KaraboEvent.ShowUnattachedSceneView: self._event_unattached_scene,
            KaraboEvent.ShowUnattachedController: self._event_controller_panel,
            KaraboEvent.OpenSceneLink: self._event_scene_link,
            KaraboEvent.RemoveProjectModelViews: self._event_remove_model_view,
            KaraboEvent.MiddlePanelClosed: self._event_middle_panel,
            KaraboEvent.ShowMacroView: self._event_show_macro,
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.CreateMainWindow: self._event_mainwindow,
            KaraboEvent.RaiseEditor: self._event_raise_editor,
            KaraboEvent.AccessLevelChanged: self._event_access_level,
            KaraboEvent.UserSession: self._event_user_session,
        }
        register_for_broadcasts(event_map)

    # -------------------------------------------------------------------
    # public interface

    def toggleReloadExtensions(self):
        """Enable the reload extensions on the mainwindow"""
        button = QPushButton(icons.refresh, "Reload Extensions")
        button.clicked.connect(self._reload_extensions)
        button.setFixedHeight(40)
        button.setVisible(True)
        self.main_window.toolbar.addWidget(button)

    def is_showing_project_item(self, model):
        """Returns True if the given `model` has a view currently showing.
        """
        return model in self._project_item_panels

    def scene_panel_data(self):
        """Return the dictionary of opened project scene panels"""
        panel_data = self._project_item_panels
        return {model: panel for model, panel in panel_data.items()
                if isinstance(model, SceneModel)}

    def use_splash_screen(self, splash):
        """Attach a QSplashScreen instance which will be closed when the first
        window appears.
        """
        if self.main_window is not None:
            splash.finish(self.main_window)

        self.splash = splash

    # --------------------  -----------------------------------------------
    # Qt callbacks

    def _event_restore_scene(self, data):
        target_window = data.get("target_window",
                                 SceneTargetWindow.MainWindow)
        model = data["model"]
        panel = self._open_scene(model, target_window, attached=True)
        if panel is not None and not panel.is_docked:
            panel.move(data["x"], data["y"])

    def _event_attached_scene(self, data):
        target_window = data.get('target_window',
                                 SceneTargetWindow.MainWindow)
        model = data['model']
        self._open_scene(model, target_window, attached=True)

    def _event_controller_panel(self, data):
        model = data["model"]
        title = f"{model.keys[0]}|Panel"
        existing = self._widget_controller_panels.items()
        existing = [panel for key, panel in existing if title == key]

        # Create a new widget controller panels and afterwards close the old
        # ones. This way the visibility counter is not disturbed!
        panel = WidgetControllerPanel(title, model)
        panel.signalPanelClosed.connect(self._on_controller_panel_close)
        for old_panel in existing:
            self._close_panel(old_panel)

        self._widget_controller_panels[title] = panel
        main_win = self.main_window
        if main_win is None:
            panel.show()
            return

        main_win.addPanel(panel, PanelAreaEnum.Middle)
        panel.onUndock()

    def _event_unattached_scene(self, data):
        target_window = data.get('target_window',
                                 SceneTargetWindow.Dialog)
        model = data['model']

        # Unattached scenes are closed if they are duplicated by simple name!
        existing = [panel for scene, panel in
                    self._unattached_scene_panels.items()
                    if scene.simple_name == model.simple_name]
        scene_panel = self._open_scene(model, target_window, attached=False)
        for panel in existing:
            self._close_panel(panel)
        position = data.get("position")
        if position is not None and target_window == SceneTargetWindow.Dialog:
            scene_panel.move(*position)

    def _event_scene_link(self, data):
        name = data.get('name', '')
        target_window = data.get('target_window')
        uuid = data.get('target')
        model = _find_scene_model(name, uuid)
        if model is not None:
            # We found the scene in our project!
            self._open_scene(model, target_window)
        else:
            position = data.get("position")
            get_db_conn().get_database_scene(
                name, uuid, target_window, position=position)

    def _event_remove_model_view(self, data):
        self._close_project_item_panels(data['models'])

    def _event_middle_panel(self, data):
        model = data.get('model')
        if model in self._project_item_panels:
            self._project_item_panels.pop(model)
        elif model in self._unattached_scene_panels:
            self._unattached_scene_panels.pop(model)
        elif model in self._widget_controller_panels:
            self._widget_controller_panels.pop(model)

        # We don't have a main window, but are either running in cinema or
        # theatre. Hence, we can close the application if no panel has been
        # opened
        if self.main_window is None and not self._unattached_scene_panels:
            close_app()

    def _event_show_macro(self, data):
        self._open_macro(data['model'])

    def _event_network(self, data):
        self.connected_to_server = data.get('status', False)
        if not self.connected_to_server:
            # Copy; _close_instance_panel will mutate
            panel_ids = list(self._instance_panels.keys())
            for inst_id in panel_ids:
                self._close_instance_panel(inst_id)
            # Close panels not associated with projects
            self._close_unattached_panels()
            if self.main_window is not None:
                icon, tooltip = get_user_session_button_data()
                self.main_window.setSessionButton(icon, tooltip)

    def _event_mainwindow(self, data):
        self._create_main_window()

    def _event_raise_editor(self, data):
        self._raise_editor(data)

    def _event_access_level(self, data):
        """React on the access level changes and modify scene and macro panel
        """
        scene_editable = access_role_allowed(AccessRole.SCENE_EDIT)
        macro_editable = access_role_allowed(AccessRole.MACRO_EDIT)
        apply_scene_edit = access_role_allowed(AccessRole.APPLY_SCENE_EDIT)
        for panel in self._project_item_panels.values():
            if isinstance(panel, ScenePanel):
                panel.setReadOnly(scene_editable)
                panel.toggleAlwaysVisibleToolbar(apply_scene_edit)
            elif isinstance(panel, MacroPanel):
                panel.setReadOnly(not macro_editable)
        for panel in self._unattached_scene_panels.values():
            panel.toggleAlwaysVisibleToolbar(apply_scene_edit)

    def _event_user_session(self, data):
        """Update the User Session button in main window and all
        scene panels, when the Temporary session state changes"""
        icon, tooltip = get_user_session_button_data()
        if self.main_window is not None:
            self.main_window.setSessionButton(icon, tooltip)
        for panel in self._project_item_panels.values():
            if isinstance(panel, ScenePanel):
                panel.setSessionButton(icon, tooltip)
        for panel in self._unattached_scene_panels.values():
            panel.setSessionButton(icon, tooltip)

    # -------------------------------------------------------------------
    # private interface

    def _raise_editor(self, data):
        """Raise the main window or editor if existent and move to cursor"""
        if self.main_window is not None:
            self.main_window.dockPanel(CONFIGURATOR_TITLE)
            self.main_window.activateWindow()
            self.main_window.raise_()
            move_to_cursor(self.main_window)
        else:
            if self._editor is None:
                self._editor = ConfigurationPanel(allow_closing=True)
                self._editor.signalPanelClosed.connect(self._editor_closed)
                self._editor.toolbar.setVisible(True)
                self._editor.resize(600, 800)
                self._editor.show()
            self._editor.activateWindow()
            self._editor.raise_()
            move_to_cursor(self._editor)

        broadcast_event(KaraboEvent.ShowConfiguration, data)

    def _open_instance_panel(self, instance_id, klass, area_enum):
        """Add a panel to main window, hide the QAction button if necessary
        """
        if instance_id in self._instance_panels:
            return
        main_win = self.main_window
        if main_win is None:
            return
        panel = klass(instance_id)
        panel.signalPanelClosed.connect(self._on_tab_close)
        main_win.addPanel(panel, area_enum)
        self._instance_panels[instance_id] = (panel, area_enum)

    def _close_instance_panel(self, instance_id):
        """Remove the panel and its linked QAction from the main window
        """
        main_win = self.main_window
        if main_win is None:
            return
        if instance_id in self._instance_panels:
            panel, area_enum = self._instance_panels.pop(instance_id)
            main_win.removePanel(panel, area_enum)

    @Slot(str)
    def _editor_closed(self, panel):
        """The dedicated editor singleton has been closed"""
        self._editor = None

    @Slot(str)
    def _on_tab_close(self, instance_id):
        """A panel emits its instance id when user close it, this function
        enables the action to reopen it
        """
        if instance_id in self._instance_panels:
            del self._instance_panels[instance_id]

    @Slot(str)
    def _on_controller_panel_close(self, panel):
        if panel in self._widget_controller_panels:
            del self._widget_controller_panels[panel]

    def _close_panel(self, panel):
        if panel is None:
            return
        if self.main_window is None:
            panel.close()
        else:
            self.main_window.removePanel(panel, PanelAreaEnum.Middle)

    def _close_project_item_panels(self, models):
        for model in models:
            self._close_panel(self._project_item_panels.get(model))

    def _close_unattached_panels(self):
        for panel in list(self._unattached_scene_panels.values()):
            self._close_panel(panel)
        for panel in list(self._widget_controller_panels.values()):
            self._close_panel(panel)

    def _create_main_window(self):
        if self.main_window is not None:
            return

        self.main_window = MainWindow()
        icon = get_application_icon()
        self.main_window.setWindowIcon(icon)
        process_qt_events(timeout=1000)

        if self.splash is not None:
            self.splash.finish(self.main_window)
            # Finish splash
            process_qt_events(timeout=1000)

        # Show main window and process events!
        self.main_window.show()
        process_qt_events(timeout=1000)
        self.main_window.activateWindow()
        process_qt_events(timeout=1000)

        check_updates = get_config()['check_updates']
        if check_updates and is_package_updated(GUI_EXTENSIONS):
            self.main_window.enableGuiExtensionUpdate()

        show_wizard = get_config()['wizard']
        if show_wizard:
            wizard = TipsTricksWizard(parent=self.main_window)
            # We can show the wizard if needed and block!
            wizard.exec()
            process_qt_events(timeout=1000)

        # Show the connection dialog after processing all pending events
        self.main_window.acServerConnect.trigger()
        if get_config()["development"]:
            self.toggleReloadExtensions()

    def _open_macro(self, model):
        if model is None:
            print("Tried to open a macro which was None!")
            return
        panel = self._project_item_panels.get(model)
        if panel is None:
            panel = MacroPanel(model)

        editable = access_role_allowed(AccessRole.MACRO_EDIT)
        panel.ui_editor.setReadOnly(not editable)
        self._show_project_item_panel(model, panel)
        send_info(type="open_macro", uuid=model.uuid, name=model.simple_name)

    def _open_scene(self, model, target_window, attached=True):
        if model is None:
            print("Tried to open a scene which was None!")
            return
        panel = self._project_item_panels.get(model)
        if panel is None:
            model.assure_svg_data()

            panel = ScenePanel(model, self.connected_to_server)
            # Set the tool color according to the defined indicators!
            karabo_topic = get_config()['broker_topic']
            panel.set_toolbar_style(karabo_topic)
            enabled = (is_authenticated() and krb_access.SERVER_READ_ONLY is
                       False)
            panel.setTemporaryButtonVisible(enabled)
            if enabled:
                # Sync the button state with current Temporary session state.
                icon, tooltip = get_user_session_button_data()
                panel.setSessionButton(icon, tooltip)

        # XXX: Only attached and access level dependent scene panels are
        # allowed to have design mode!
        editable = access_role_allowed(AccessRole.SCENE_EDIT)
        if not attached:
            panel.setUnattached(editable)
        else:
            panel.setReadOnly(editable)

        # The apply scene edit is global setting
        apply_scene_edit = access_role_allowed(AccessRole.APPLY_SCENE_EDIT)
        panel.toggleAlwaysVisibleToolbar(apply_scene_edit)

        self._show_project_item_panel(model, panel, attached)

        send_info(type="open_scene", uuid=model.uuid,
                  name=model.simple_name, attached=attached)
        if self.main_window is None:
            return panel

        if target_window is SceneTargetWindow.MainWindow:
            panel.onDock()
        elif target_window is SceneTargetWindow.Dialog:
            panel.onUndock()
        return panel

    def _show_project_item_panel(self, model, panel, attached=True):
        has_panel = model in self._project_item_panels

        if attached:
            self._project_item_panels[model] = panel
        else:
            self._unattached_scene_panels[model] = panel

        if self.splash is not None:
            self.splash.close()
            self.splash = None

        if self.main_window is None:
            panel.show()
            return
        if not has_panel:
            self.main_window.addPanel(panel, PanelAreaEnum.Middle)
        self.main_window.selectPanel(panel, PanelAreaEnum.Middle)

    @Slot()
    def _reload_extensions(self):
        """Request to reload the gui extensions and close all scenes"""
        get_logger().info("Reloading the GUI Extensions ...")
        load_extensions()
        panels = [p for p in self._project_item_panels.values()
                  if isinstance(p, ScenePanel)]
        for panel in panels:
            self._close_panel(panel)
        # And all unattached scene panels
        self._close_unattached_panels()


def _find_scene_model(name, uuid):
    """Find a SceneModel which is already open in the project.
    """

    class _Visitor:
        found = None

        def __call__(self, obj):
            if isinstance(obj, SceneModel):
                if obj.uuid == uuid:
                    self.found = obj
                    return True

    project = get_project_model().root_model
    if project is None:
        return None

    visitor = _Visitor()
    walk_traits_object(project, visitor, fast_exit=True)

    if visitor.found is None:
        msg = ("The UUID of the linked scene <b>{}</b> was not found in the "
               "project! Trying to retrieve the scene in from the project "
               "<b>database</b> instead.".format(name))
        get_logger().warning(msg)
        return None

    return visitor.found



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 7, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import math
from contextlib import contextmanager

from qtpy.QtCore import QPoint
from qtpy.QtGui import QFontMetrics

from karabogui.fonts import get_qfont

from .const import GRID_STEP, SCREEN_MAX_VALUE


def calc_bounding_rect(collection):
    """Compute the bounding rectangle for a collection of objects.

    Each object in the collection must have a geometry method which returns a
    QRect.
    """
    left, top, right, bottom = SCREEN_MAX_VALUE, SCREEN_MAX_VALUE, 0, 0
    for item in collection:
        rect = item.geometry()
        left = min(left, rect.left())
        top = min(top, rect.top())

        # Qt right and bottom are always width - 1 and height - 1,
        # respectively. Therefore, we correct it.
        right = max(right, rect.right() + 1)
        bottom = max(bottom, rect.bottom() + 1)

    # Return an empty rect when nothing is there
    if left == SCREEN_MAX_VALUE and top == SCREEN_MAX_VALUE:
        return (0, 0, 0, 0)

    # Return x, y, width, height
    return (left, top, right - left, bottom - top)


def calc_rect_from_text(font, text):
    """Compute the rectangle to fit the given ``text``.

    A tuple including x, y, width and height.
    """
    q_font = get_qfont(font)
    fm = QFontMetrics(q_font)
    CONTENT_MARGIN = 10
    width = fm.width(text) + CONTENT_MARGIN
    height = fm.height() + CONTENT_MARGIN

    return (0, 0, width, height)


@contextmanager
def save_painter_state(painter):
    painter.save()
    yield
    painter.restore()


def round_down_to_grid(num):
    """Round down to the nearest grid step"""
    return math.floor(num / GRID_STEP) * GRID_STEP


def round_up_to_grid(num):
    """Round up to the nearest grid step"""
    return math.ceil(num / GRID_STEP) * GRID_STEP


def calc_snap_pos(pos):
    """Calculate the snap position by rounding down the x-y points to the
       nearest grid"""
    return QPoint(round_down_to_grid(pos.x()), round_down_to_grid(pos.y()))


def calc_rotated_point(x, y, angle=0, scale=1):
    """Rotate a point with the angle and the scale.
       Angle should be in degrees."""

    radians = math.radians(angle)
    cos, sin = math.cos(radians), math.sin(radians)

    rotated_x = x * cos + y * sin
    rotated_y = - x * sin + y * cos

    return QPoint(int(rotated_x * scale), int(rotated_y * scale))


def add_offset(model, x=0, y=0):
    # Check if the models has x or y, as there are some that do not follow
    # the item interface. We do nothing instead.
    trait_names = model.copyable_trait_names()
    if "x" not in trait_names or "y" not in trait_names:
        return

    model.trait_setq(x=model.x + x, y=model.y + y)

    # Recursively offset the children
    if "children" in model.trait_names():
        for child in model.children:
            add_offset(child, x=x, y=y)


def calc_relative_pos(models):
    x, y = (SCREEN_MAX_VALUE, SCREEN_MAX_VALUE)
    for model in models:
        x = min(x, model.x)
        y = min(y, model.y)
    return x, y



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import Point
from qtpy.QtCore import QRectF, QSizeF

from .base import KaraboROI
from .utils import ImageRegion


class RectROI(KaraboROI):

    def _add_handles(self):
        self.addScaleHandle([1, 1], [0, 0])

    def get_region(self, imageItem):
        # We only get the region if the ROI is inside the image
        if not imageItem.rect().intersects(self.rect()):
            return ImageRegion()

        # Compute ROI manually
        start_x, start_y = self._absolute_position
        if start_x < 0:
            start_x = 0
        if start_y < 0:
            start_y = 0

        end_x, end_y = self._absolute_position + self._absolute_size

        _, _, max_x, max_y = imageItem.boundingRect().getCoords()
        if end_x > max_x:
            end_x = max_x
        if end_y > max_y:
            end_y = max_y

        x_slice = slice(int(start_x), int(end_x))
        y_slice = slice(int(start_y), int(end_y))

        region = imageItem.image[y_slice, x_slice]

        image_region = ImageRegion(region, ImageRegion.Area, x_slice, y_slice)

        return image_region

    # ---------------------------------------------------------------------
    # PyQt methods

    def setRect(self, rect):
        self.setPos(rect.topLeft(), finish=False)
        self.setSize(rect.size(), finish=False)

    def rect(self):
        return QRectF(self.pos(), QSizeF(*self.size()))

    # ---------------------------------------------------------------------
    # PyQtGraph methods

    def getSnapPosition(self, pos, snap=None):
        """Patching to support snapping based on scaling"""
        scaled_pos = np.floor(np.array(pos) / self._scaling) * self._scaling

        return Point(*scaled_pos)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget
from traits.api import Enum, Instance, Int, WeakRef

from karabo.common.scenemodel.api import (
    DetectorGraphModel, build_graph_config, restore_graph_config)
from karabo.native import Encoding, Timestamp
from karabogui.binding.api import ImageBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.common.api import AuxPlots, Axes
from karabogui.graph.image.api import (
    KaraboImageNode, KaraboImagePlot, KaraboImageView, karabo_default_image)
from karabogui.util import SignalBlocker

from .util import get_ui_file


class FrameSlider(QWidget):
    axisChanged = Signal(str)
    cellChanged = Signal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        ui_file = get_ui_file("frame_slider.ui")
        uic.loadUi(ui_file, self)

        # Populate our axis combobox
        self.cb_axis.addItems([Axes.X.name, Axes.Y.name, Axes.Z.name])

        self.cb_axis.currentIndexChanged[str].connect(self.axisChanged.emit)
        self.sl_cell.valueChanged.connect(self.on_slider_moved)
        self.sb_cell.valueChanged.connect(self.on_value_changed)

    def on_slider_moved(self, value):
        with SignalBlocker(self.sb_cell):
            self.sb_cell.setValue(value)
        self.cellChanged.emit(value)

    def on_value_changed(self, value):
        with SignalBlocker(self.sl_cell):
            self.sl_cell.setSliderPosition(value)
        self.cellChanged.emit(value)

    def set_slider_maximum(self, max_value):
        """Set the maximum possible value and reset the current position

        :param int max_value:
        """
        self.sl_cell.setMaximum(max_value)
        self.sb_cell.setMaximum(max_value)

    def set_axis(self, axis):
        """Sets the combo box axis based on the given axis name

        :param axis: the selected axes
        :type axis: Enum (Axes)
        """
        self.cb_axis.setCurrentIndex(Axes(axis).value)

    def set_cell(self, cell):
        self.sb_cell.setValue(cell)

    def reset(self):
        spinbox_blocker = SignalBlocker(self.sb_cell)
        slider_blocker = SignalBlocker(self.sl_cell)

        with spinbox_blocker, slider_blocker:
            self.sl_cell.setSliderPosition(0)
            self.sb_cell.setValue(0)
        self.cellChanged.emit(0)


@register_binding_controller(ui_name="Detector Graph",
                             klassname="DetectorGraph",
                             binding_type=ImageBinding, priority=10,
                             can_show_nothing=False)
class DisplayDetectorGraph(BaseBindingController):
    model = Instance(DetectorGraphModel, args=())

    # internal traits
    _axis = Enum(*Axes)
    _cell = Int(0)

    _frame_slider = Instance(FrameSlider)
    _image_node = Instance(KaraboImageNode, args=())
    _plot = WeakRef(KaraboImagePlot)

    def create_widget(self, parent):
        """
        Setups the Detector Graph Image widget. Differently from the Graph
        Widget, this one has an additional frame on the right side which is
        responsible for controlling which pulse/stack is shown.
        """
        # Setup parent widget
        widget = KaraboImageView(parent=parent)
        widget.stateChanged.connect(self._change_model)
        widget.toolTipChanged.connect(self.show_timestamp_tooltip)
        widget.add_picker()
        widget.add_roi()
        widget.add_colorbar()
        widget.add_aux(AuxPlots.ProfilePlot, smooth=True)
        widget.add_aux(AuxPlots.Histogram)
        # Finalize
        widget.add_toolbar()

        self._frame_slider = FrameSlider(parent=widget)
        self._frame_slider.axisChanged.connect(self._axis_changed)
        self._frame_slider.cellChanged.connect(self._cell_changed)
        widget.add_widget(self._frame_slider, row=0, col=2)
        self._frame_slider.setVisible(False)

        # Get a reference for our plotting
        self._plot = widget.plot()

        # QActions
        widget.add_axes_labels_dialog()
        widget.add_transforms_dialog()

        # Default axes
        self._axis = Axes.Z

        # Restore the model information
        widget.restore(build_graph_config(self.model))

        self._frame_slider.set_axis(self._axis)
        self._frame_slider.set_cell(self._cell)

        return widget

    def clear_widget(self):
        self._plot.setData(karabo_default_image())

    # -----------------------------------------------------------------------
    # Qt Slots

    def show_timestamp_tooltip(self):
        image_node = self.proxy.value
        if image_node is None or getattr(image_node, "pixels", None) is None:
            # No schema for output channel, pixels not available
            return
        timestamp = image_node.pixels.value.data.timestamp
        if timestamp is None:
            # Can happen when we toggle aux plots
            return
        diff = Timestamp().toTimestamp() - timestamp.toTimestamp()
        self.widget.setToolTip("{} --- Last image received {:.3f} s "
                               "ago".format(self.proxy.key, diff))

    def _axis_changed(self, axis):
        """Called whenever an axis is selected in the FrameSlider

        :param axis: the selected axis
        :type axis: basestring
        """
        if self._image_node is None:
            return

        self._axis = Axes[axis]
        self._set_slider_max(self._image_node.get_axis_dimension(self._axis))
        self._frame_slider.reset()

    def _cell_changed(self, value):
        if self._image_node is None:
            return

        self._cell = value
        self._update_image()

    def _update_image(self):
        # Get region to plot
        np_image = self._image_node.get_slice(self._axis, self._cell)
        self._plot.set_image(np_image)

    def _change_model(self, content):
        self.model.trait_set(**restore_graph_config(content))

    # -----------------------------------------------------------------------

    def value_update(self, proxy):
        self._image_node.set_value(proxy.value)

        if not self._image_node.is_valid:
            return

        # Hide the slider when there's no multiple images
        self._frame_slider.setVisible(self._image_node.dim_z != 0)

        if self._image_node.encoding == Encoding.GRAY:
            slider_max = self._image_node.get_axis_dimension(self._axis)
            self._set_slider_max(slider_max)

        self._update_image()

    def _set_slider_max(self, value):
        max_value = max(value - 1, 0)
        self._frame_slider.set_slider_maximum(max_value)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import datetime

import numpy as np

from karabo.native import Hash, Timestamp


def build_historic_data_float():
    """build an array of historic data with time stamps"""
    # 10 random values in between [0, 100]
    vals = np.random.rand(10) * 100.

    # use now as base, add random values in minutes to the past
    now = datetime.datetime.now()
    timefmt = r"%Y-%m-%d %H:%M:%S"

    def time_rewind(nmin):
        dt = now + datetime.timedelta(minutes=-nmin)
        return Timestamp(dt.strftime(timefmt)).toDict()

    tss = map(time_rewind, reversed(sorted(vals)))
    ret = []
    for vv, tt in zip(vals, tss):
        h = Hash('v', vv)
        for attk, attv in tt.items():
            h.setAttribute('v', attk, attv)
        ret.append(h)
    return ret


def _build_value_string(vals):
    now = datetime.datetime.now()
    timefmt = r"%Y-%m-%d %H:%M:%S"

    ret = []
    for vv in vals:
        dt = now + datetime.timedelta(minutes=-1)
        tt = Timestamp(dt.strftime(timefmt)).toDict()
        h = Hash('v', vv)
        for attk, attv in tt.items():
            h.setAttribute('v', attk, attv)
        ret.append(h)
    return ret


def build_historic_state_string():
    """build an array of historic data state string with time stamps"""
    vals = ['ON', 'OFF', 'ACQUIRING', 'STOPPED', 'DISABLED', 'RUNNING',
            'CHANGING', 'ERROR', 'UNKNOWN', 'STARTED', 'MOVING']
    return _build_value_string(vals)


def build_historic_alarm_string():
    """build an array of historic data alarm string with time stamps"""
    vals = ['none', 'warn', 'warnLow', 'warnHigh', 'warnVarianceLow',
            'warnVarianceHigh', 'alarm', 'alarmLow', 'alarmHigh',
            'alarmVarianceLow', 'alarmVarianceHigh', 'interlock']
    return _build_value_string(vals)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QItemSelectionModel

from karabogui.testing import click_button, singletons, system_hash
from karabogui.topology.api import SystemTopology

from ..topology_device_dialog import TopologyDeviceDialog


def test_device_dialog(gui_app):
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        dialog = TopologyDeviceDialog()
        assert dialog.device_id == ""
        assert dialog.filter_model.rowCount() == 2
        selection = dialog.ui_devices_view.selectionModel()
        model = dialog.filter_model
        index = model.index(0, 0)
        selection.setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)
        dialog.done(1)
        assert dialog.device_id == "divvy"

        dialog.filter_model.setFilterFixedString("nooo")
        assert dialog.filter_model.rowCount() == 0
        # Clear to get previous selection
        click_button(dialog.ui_clear)
        assert dialog.filter_model.rowCount() == 2
        dialog.ui_interface.setCurrentIndex(1)
        # Interface, all filtered
        assert dialog.filter_model.rowCount() == 0
        click_button(dialog.ui_clear)
        assert dialog.filter_model.rowCount() == 2



from karabo.common.scenemodel.api import PopupButtonModel
from karabogui.dialogs.api import PopupButtonDialog

TEXT = "XFEL: the world's largest X-ray laser."


def test_pop_button_dialog(gui_app):
    model = PopupButtonModel(
        text=TEXT, x=0, y=0, width=100, height=100)
    dialog = PopupButtonDialog(model)
    assert dialog.leText.toPlainText() == model.text



#############################################################################
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QAbstractTableModel, QMimeData, QModelIndex, Qt
from qtpy.QtGui import QBrush, QColor

from karabo.native import AccessMode, Hash
from karabogui.binding.api import BoolBinding, VectorBinding, get_default_value
from karabogui.indicators import get_state_color
from karabogui.util import create_table_string

from .utils import (
    create_mime_data, is_state_display_type, list2string, quick_table_copy,
    string2list)


class TableModel(QAbstractTableModel):
    def __init__(self, binding, set_edit_value, parent=None):
        super().__init__(parent)
        self._set_edit_value = set_edit_value
        self._readonly = False
        self._data = []

        # Set the bindings and the default row hash
        bindings = binding.bindings
        header = list(bindings.keys())
        default_row = Hash()
        for key in header:
            binding = bindings[key]
            value = get_default_value(binding, force=True)
            value = binding.validate_trait("value", value)
            default_row[key] = value

        self._bindings = bindings
        self._header = header
        self.default_row = default_row

    # Qt Methods
    # ----------------------------------------------------------------------

    def rowCount(self, parent=None):
        """Reimplemented function of QAbstractTableModel"""
        return len(self._data)

    def columnCount(self, parent=None):
        """Reimplemented function of QAbstractTableModel"""
        return len(self._header)

    def data(self, index, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractTableModel"""
        if not index.isValid():
            return None

        row, column = index.row(), index.column()
        if role == Qt.CheckStateRole:
            key = self._header[column]
            value = self._data[row][key]
            binding = self._bindings[key]
            if isinstance(binding, BoolBinding):
                return Qt.Checked if value else Qt.Unchecked
        elif role in (Qt.DisplayRole, Qt.ToolTipRole, Qt.EditRole):
            key = self._header[column]
            value = self._data[row][key]
            binding = self._bindings[key]
            if isinstance(binding, VectorBinding):
                value = list2string(value)
            return str(value)
        elif role == Qt.BackgroundRole:
            key = self._header[column]
            if is_state_display_type(self._bindings[key]):
                value = self._data[row][key]
                try:
                    color = get_state_color(value)
                except ValueError:
                    return None
                return QBrush(QColor(*color))

        return None

    def _build_tooltip(self, key, attributes):
        """Build a tooltip according to the attributes"""
        selection = ["displayedName", "defaultValue", "valueType",
                     "unitSymbol", "metricPrefixSymbol",
                     "minInc", "maxInc", "minExc", "maxExc", "options"]
        info = {"key": key}
        for akey in selection:
            avalue = attributes.get(akey)
            if avalue is not None:
                info.update({akey: str(avalue)})

        table = create_table_string(info)
        if description := attributes.get("description"):
            header = f"<center><i>{description}</i></center><hr>"
            table = header + table
        return table

    def headerData(self, section, orientation, role):
        """Reimplemented function of QAbstractTableModel"""
        if role == Qt.DisplayRole:
            if orientation == Qt.Vertical:
                return str(section)
            elif orientation == Qt.Horizontal:
                key = self._header[section]
                binding = self._bindings[key]
                units = binding.unit_label
                # Make sure to always show a header!
                displayed = binding.displayedName or key

                return f"{displayed} [{units}]" if units else displayed

        elif role == Qt.ToolTipRole and orientation == Qt.Horizontal:
            key = self._header[section]
            attrs = self._bindings[key].attributes
            return self._build_tooltip(key=key, attributes=attrs)

        return None

    def flags(self, index):
        """Reimplemented function of QAbstractTableModel"""
        if not index.isValid():
            return Qt.NoItemFlags

        flags = (Qt.ItemIsEnabled | Qt.ItemIsSelectable |
                 Qt.ItemIsDragEnabled | Qt.ItemIsEditable |
                 Qt.ItemNeverHasChildren)
        key = self._header[index.column()]
        # Get an enum for the AccessMode
        binding = self._bindings[key]
        accessMode = binding.accessMode
        if isinstance(binding, BoolBinding) and not self._readonly:
            flags &= ~Qt.ItemIsEditable
            if accessMode is AccessMode.RECONFIGURABLE:
                flags |= Qt.ItemIsUserCheckable
        elif accessMode is AccessMode.READONLY:
            flags &= ~Qt.ItemIsEditable

        return flags

    def setData(self, index, value, role=Qt.EditRole):
        """Reimplemented function of QAbstractTableModel"""
        if not index.isValid():
            return False

        row, column = index.row(), index.column()
        if role == Qt.CheckStateRole:
            key = self._header[column]
            binding = self._bindings[key]
            if isinstance(binding, BoolBinding):
                value = True if value == Qt.Checked else False
                self._data[row][key] = value
                self.dataChanged.emit(index, index)
                self._set_edit_value(self._data)
                return True

            return False

        elif role in (Qt.DisplayRole, Qt.EditRole):
            key = self._header[column]
            binding = self._bindings[key]
            if isinstance(binding, VectorBinding):
                value = string2list(value)
            value = binding.validate_trait("value", value)
            self._data[row][key] = value
            self.dataChanged.emit(index, index)
            self._set_edit_value(self._data)
            return True

        return False

    def mimeData(self, indices):
        """Reimplemented function of QAbstractTableModel"""
        row = indices[0].row()
        mime = QMimeData()
        mime.setData("tableData", create_mime_data(row=row))
        return mime

    def updateData(self, data):
        """External quick update of the table data"""
        roles = [Qt.BackgroundRole, Qt.CheckStateRole, Qt.DisplayRole,
                 Qt.ToolTipRole]
        rows = self.rowCount() - 1
        columns = self.columnCount() - 1
        for row, hsh in enumerate(data):
            self._data[row].update(hsh)

        first_index = self.index(0, 0)
        last_index = self.index(rows, columns)
        self.dataChanged.emit(first_index, last_index, roles)

    def moveRow(self, source_parent, source_row, parent, row):
        """Reimplemented function of QAbstractTableModel"""
        source_last = source_row - 1
        self.beginMoveRows(source_parent, source_row, source_last, parent, row)
        try:
            hsh = self._data.pop(source_row)
            self._data.insert(row, hsh)
        finally:
            self.endMoveRows()
            self._set_edit_value(self._data)

    def insertRows(self, pos, rows, index, *,
                   copy_row=None, from_device=False):
        """Reimplemented function of QAbstractTableModel"""
        self.beginInsertRows(QModelIndex(), pos, pos + rows - 1)
        try:
            if copy_row is None:
                copy_row = self.default_row
            for row_nr in range(rows):
                column_hash = quick_table_copy(copy_row)
                if pos + row_nr < len(self._data):
                    self._data.insert(pos + row_nr, column_hash)
                else:
                    self._data.append(column_hash)
        finally:
            self.endInsertRows()

        if not from_device:
            self._set_edit_value(self._data)
        return True

    def removeRows(self, pos, rows, index, *, from_device=False):
        """Reimplemented function of QAbstractTableModel"""
        self.beginRemoveRows(QModelIndex(), pos, pos + rows - 1)
        try:
            for row_nr in range(rows, 0, -1):
                self._data.pop(pos + row_nr - 1)
        finally:
            self.endRemoveRows()
        if not from_device:
            self._set_edit_value(self._data)

        return True

    # Public Interface
    # -----------------------------------------------------------------------

    def index_ref(self, index):
        return index

    def get_model_data(self, row, column):
        """Retrieve the validated model data for `row` and `column`"""
        key = self._header[column]
        binding = self._bindings[key]

        if isinstance(binding, BoolBinding):
            value = self.index(row, column).data(
                role=Qt.CheckStateRole) == Qt.Checked
        else:
            value = self.index(row, column).data(role=Qt.DisplayRole)
            if isinstance(binding, VectorBinding):
                value = string2list(value)
            value = binding.validate_trait("value", value)
        return key, value

    def get_header_key(self, section):
        """Retrieve the key of the binding for the section"""
        return self._header[section]

    def clear_model(self):
        """Clear the model and remove all data"""
        self.beginResetModel()
        try:
            self._data = []
        finally:
            self.endResetModel()

    def set_readonly(self, value):
        """Set the readonly role of the table element"""
        self._readonly = value

    # ------------------------------------------------------------------------
    # Additional methods

    def duplicate_row(self, row):
        """Duplicate a row element at `row` in the table"""
        copy_row = self._data[row]
        self.add_row_below(row, copy_row=copy_row)

    def add_row(self, row, copy_row=None):
        """Insert a row element `copy_row` at `row` in the table"""
        self.insertRows(row, 1, QModelIndex(), copy_row=copy_row)

    def add_row_below(self, row, copy_row=None):
        """Add a row element `copy_row` below `row` in the table"""
        self.add_row(row + 1, copy_row=copy_row)

    def remove_row(self, row):
        """Remove a `row` element from the table"""
        self.removeRows(row, 1, QModelIndex())

    def move_row_up(self, row):
        """In a move step process move a row element up"""
        if row > 0:
            self.moveRow(QModelIndex(), row, QModelIndex(), row - 1)

    def move_row_down(self, row):
        """In a move process move a row element down"""
        if row < self.rowCount() - 1:
            self.moveRow(QModelIndex(), row, QModelIndex(), row + 1)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from karabo.common.scenemodel.api import read_scene
from karabogui.sceneview.api import SceneView

DATA_DIR = Path(__file__).parent.joinpath("data")


def _iter_data_files(directory):
    for fn in Path.iterdir(directory):
        if fn.suffix == ".svg":
            yield Path(directory).joinpath(fn)


def test_reading_scenes(gui_app):
    for fn in _iter_data_files(DATA_DIR):
        model = read_scene(fn)
        view = SceneView(model)
        view.show()
        assert view is not None
        view.destroy()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QModelIndex, Slot
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QHeaderView, QTableView

from karabo.native import decodeBinary


class KaraboTableView(QTableView):
    """The KaraboTableView supports internal `Drag and Drop` operations"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._readonly = True
        header = self.horizontalHeader()
        header.sectionDoubleClicked.connect(self._header_resize)

    def set_read_only(self, ro):
        """Public method to set the readonly status `ro` of the table view

        Note: Method added with Karabo 2.16.X
        """
        self._readonly = ro
        writable = not ro
        self.setAcceptDrops(writable)
        self.setDragEnabled(writable)
        self.setDropIndicatorShown(writable)

    def currentIndex(self):
        """Convenience function to return the current reference index

        Note: Method added with Karabo 2.16.X
        """
        selection = self.selectionModel()
        if selection is not None and selection.hasSelection():
            return selection.currentIndex()
        return QModelIndex()

    def keyPressEvent(self, event):
        """Reimplemented method of `QTableView`"""
        if self._readonly:
            return super().keyPressEvent(event)

        if event.matches(QKeySequence.New):
            model = self.model()
            index = model.index_ref(self.currentIndex())
            if not index.isValid():
                index = model.index(0, 0)
            model.add_row(index.row() + 1)
            self.selectRow(self.currentIndex().row() + 1)
            event.accept()
            return
        elif event.matches(QKeySequence.Delete):
            model = self.model()
            index = model.index_ref(self.currentIndex())
            if index.isValid():
                model.remove_row(index.row())
            event.accept()
            return
        elif event.matches(QKeySequence.SelectPreviousLine):
            model = self.model()
            index = model.index_ref(self.currentIndex())
            if index.isValid():
                model.move_row_up(index.row())
                self.selectRow(self.currentIndex().row() - 1)
            event.accept()
            return
        elif event.matches(QKeySequence.SelectNextLine):
            model = self.model()
            index = model.index_ref(self.currentIndex())
            if index.isValid():
                model.move_row_down(index.row())
                self.selectRow(self.currentIndex().row() + 1)
            event.accept()
            return
        return super().keyPressEvent(event)

    @Slot()
    def _header_resize(self):
        """Resize the table view to the model contents"""
        model = self.model()
        if model is not None and model.columnCount() > 1:
            columns = model.columnCount() - 1
            header = self.horizontalHeader()
            header.resizeSections(QHeaderView.ResizeToContents)
            header.resizeSection(columns, QHeaderView.Stretch)

    def dragEnterEvent(self, event):
        """Reimplemented method of `QTableView`"""
        self._inspect_event(event)

    def dragMoveEvent(self, event):
        """Reimplemented method of `QTableView`"""
        self._inspect_event(event)

    def dropEvent(self, event):
        """Reimplemented method of `QTableView`"""
        mime = self._inspect_event(event)
        source_row, index = mime
        model = self.model()
        source_index = model.index_ref(index)
        row = (source_index.row() if source_index.isValid()
               else model.rowCount() - 1)
        if row == source_row:
            return
        self.model().moveRow(QModelIndex(), source_row,
                             QModelIndex(), row)
        self.selectRow(index.row())

    def _inspect_event(self, event):
        """Inspect the drag and either ignore or accept and provide data"""
        if event.source() is not self:
            event.ignore()
            return

        model = self.model()
        if model is None:
            event.ignore()
            return

        mime = event.mimeData().data("tableData").data()
        if not mime:
            event.ignore()
            return

        source_row = decodeBinary(mime)["row"]
        index = self.indexAt(event.pos())
        event.accept()
        return source_row, index



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import json

from qtpy.QtCore import QMimeData, Qt
from qtpy.QtGui import QColor, QPalette
from qtpy.QtWidgets import QStyle

from karabo.common.api import State
from karabo.native import AccessMode, Assignment
from karabogui import icons
from karabogui.binding.api import (
    BindingRoot, BoolBinding, CharBinding, FloatBinding, ImageBinding,
    IntBinding, NodeBinding, StringBinding, VectorHashBinding,
    WidgetNodeBinding, get_binding_value, get_editor_value)
from karabogui.controllers.api import get_compatible_controllers
from karabogui.indicators import (
    ERROR_COLOR_ALPHA, OK_COLOR, PROPERTY_READONLY_COLOR, UNKNOWN_COLOR_ALPHA)
from karabogui.itemtypes import ConfiguratorItemType

# The fixed height of rows in the configurator
FIXED_ROW_HEIGHT = 30
RECURSIVE_BINDING = (BindingRoot, NodeBinding)
_EMPTY_VALUE_BINDINGS = (BindingRoot, NodeBinding, VectorHashBinding)


class ButtonState:
    PRESSED = QStyle.State_Enabled | QStyle.State_Sunken
    ENABLED = QStyle.State_Enabled | QStyle.State_Raised | QStyle.State_Off
    DISABLED = QStyle.State_On


def _get_item_type(binding):
    """Get the item type for a `binding`

    We can only drag nodes and leaf elements
    """
    if isinstance(binding, NodeBinding):
        return ConfiguratorItemType.NODE
    return ConfiguratorItemType.LEAF


def dragged_configurator_items(proxies):
    """Create a QMimeData object containing items dragged from the configurator
    """
    dragged = []
    for proxy in proxies:
        if proxy.binding is None:
            continue

        # Collect the relevant information
        binding = proxy.binding
        item_type = _get_item_type(binding)
        default_name = proxy.path.split('.')[-1]
        data = {
            'key': proxy.key,
            'label': binding.displayedName or default_name,
            'type': item_type,
        }

        factories = get_compatible_controllers(binding, can_edit=False)
        if factories:
            data['display_widget_class'] = factories[0].__name__
        if binding.accessMode is AccessMode.RECONFIGURABLE:
            factories = get_compatible_controllers(binding, can_edit=True)
            if factories:
                data['edit_widget_class'] = factories[0].__name__
        # Add it to the list of dragged items
        dragged.append(data)

    if not dragged:
        return None

    mimeData = QMimeData()
    mimeData.setData('source_type', bytearray('ParameterTreeWidget',
                                              encoding='UTF-8'))
    mimeData.setData('tree_items', bytearray(json.dumps(dragged),
                                             encoding='UTF-8'))
    return mimeData


def get_child_names(proxy, level):
    """Return all the names of a proxy's editable children."""
    binding = proxy.binding
    if not isinstance(binding, RECURSIVE_BINDING):
        return []

    ret = binding.children_names.get(level, None)
    # lazily cache visible children names
    if ret is None:
        ret = [name for name in binding.value if
               getattr(binding.value, name).requiredAccessLevel <= level]
        binding.children_names[level] = ret
    return ret


def is_mandatory(binding):
    """Retrieves if a binding or a binding within a node is mandatory

    This function is solely used for DeviceClass proxies
    """
    if binding.assignment is Assignment.MANDATORY:
        return True
    elif isinstance(binding, RECURSIVE_BINDING):
        ret = False

        def recurse(binding):
            """Walk through a node binding for mandatory parameters
            """
            nonlocal ret
            for name in binding.value:
                node = getattr(binding.value, name)
                if isinstance(node, RECURSIVE_BINDING):
                    recurse(node)
                elif node.assignment is Assignment.MANDATORY:
                    ret = True
                    break

        recurse(binding)
        return ret

    return False


def get_device_state_string(device_proxy):
    """Return a device state as a STRING to be checked with
    BaseBinding.is_allowed.

    NOTE: BaseBinding.is_allowed is tolerant of empty string states because it
    does not try to make the passed in string a `State` instance. This is an
    important property!
    """
    state_binding = device_proxy.state_binding
    if state_binding is None:
        return ''
    return get_binding_value(state_binding, '')


def get_device_locked_string(device_proxy):
    """Return a check if a device proxy is locked!
    """
    locked_binding = device_proxy.locked_binding
    if locked_binding is None:
        return ''
    return get_binding_value(locked_binding, '')


def get_icon(binding):
    """Get the proper icon to show next to a property in the configurator
    """
    if len(binding.options) > 0:
        return icons.enum

    icon = icons.undefined
    if isinstance(binding, (CharBinding, StringBinding)):
        icon = icons.string
    elif isinstance(binding, IntBinding):
        icon = icons.int
    elif isinstance(binding, FloatBinding):
        icon = icons.float
    elif isinstance(binding, BoolBinding):
        icon = icons.boolean
    elif isinstance(binding, (ImageBinding, WidgetNodeBinding)):
        icon = icons.image

    return icon


def get_proxy_value(index, proxy, is_edit_col=False):
    """Return the actual value of the given `proxy`, depending on whether this
    is requested for an editable column
    """
    is_editable = index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable
    if is_edit_col and not is_editable:
        return ''

    binding = proxy.binding
    if isinstance(binding, _EMPTY_VALUE_BINDINGS):
        return ''

    value = _proxy_value(proxy, is_edit_col)
    if isinstance(value, (bytes, bytearray)):
        return ''

    return value


def get_proxy_value_offline(proxy, is_edit_col=False):
    """Return the offline proxy value"""
    binding = proxy.binding
    if (binding.accessMode is AccessMode.READONLY or
            isinstance(binding, _EMPTY_VALUE_BINDINGS)):
        return ''

    value = _proxy_value(proxy, is_edit_col)
    if isinstance(value, (bytes, bytearray)):
        return ''

    units = binding.unit_label
    value = str(value)
    return value + (' ' + units if units and value else '')


def get_proxy_color_offline(index, proxy):
    """Return an appropriate color for an `index` and `proxy`"""
    if not index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable:
        # Account global access level
        return PROPERTY_READONLY_COLOR

    binding = proxy.binding
    if (binding.accessMode is AccessMode.READONLY
            or binding.assignment is Assignment.INTERNAL):
        return PROPERTY_READONLY_COLOR
    return None


def handle_default_state(allowed, state):
    """Determine the resting state of a given box's button.
    """
    if allowed and state != ButtonState.PRESSED:
        state = ButtonState.ENABLED
    if not allowed:
        state = ButtonState.DISABLED
    return state


def set_fill_rect(painter, option, index):
    """Update the rectangle of the given `painter` depending on the given
    `options`
    """
    if option.state & QStyle.State_Selected:
        if option.state & QStyle.State_Active:
            painter.fillRect(option.rect, option.palette.highlight())
        elif not (option.state & QStyle.State_HasFocus):
            # XXX: Palette background is deprecated in Qt5. Take the brush
            # with background role
            painter.fillRect(option.rect,
                             option.palette.brush(QPalette.Background))
    else:
        brush = index.data(Qt.BackgroundRole)
        if brush is not None:
            painter.fillRect(option.rect, brush)


# ----------------------------------------------------------------------------
# Private details


def _proxy_value(proxy, is_edit_col):
    """If a value is needed from the editable column, then get the value from
    the binding. This is in case the user has made a change which was not
    yet applied.
    """
    if is_edit_col:
        return get_editor_value(proxy, '')
    return get_binding_value(proxy, '')


def threshold_triggered(value, limit_low, limit_high):
    """Check if a value exceeds the limits low or high

    This method is typically used to check alarms and warnings
    """
    if ((limit_low is not None and value < limit_low) or
            (limit_high is not None and value > limit_high)):
        return True
    return False


_STATE_QCOLOR_ALPHA = {
    State.ERROR.name: QColor(*ERROR_COLOR_ALPHA),
    State.UNKNOWN.name: QColor(*UNKNOWN_COLOR_ALPHA)
}

_STATE_QCOLOR_DEFAULT = QColor(*OK_COLOR)


def get_qcolor_state(state):
    """Get the opague background color according to a `state` string
    for the configurator
    """
    return _STATE_QCOLOR_ALPHA.get(state, _STATE_QCOLOR_DEFAULT)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import GraphicsObject, Point, functions as fn, getConfigOption
from qtpy.QtCore import QPointF, QRectF, Qt, Signal, Slot
from qtpy.QtGui import QColor, QImage, QTransform, qRgb
from scipy.ndimage import zoom

from karabogui.graph.common.api import MouseTool

from .utils import bytescale, map_rect_to_transform

DIMENSION_DOWNSAMPLE = [(500, 1.5), (1000, 2)]  # [(dimension, min downsample)]
NULL_COLOR = QColor(255, 255, 255, 70)


def karabo_default_image():
    return np.zeros(shape=(10, 10), dtype=np.uint8)


class KaraboImageItem(GraphicsObject):
    clicked = Signal(float, float)
    hovered = Signal(object, object)

    sigImageChanged = Signal()

    def __init__(self, parent=None):
        super().__init__()
        self.auto_levels = True

        self._origin = np.array([0, 0])
        self._rect = None
        self.autoDownsample = True

        self.downsample_order = 0  # nearest-neighbors downsampling
        self._qlut = None
        self._lastDownsample = None
        self._slice_rect = None
        self._view_rect = None
        self.lut = None
        self.image = None
        self.qimage = None

        self.clickTools = {MouseTool.Picker}
        self.levels = None
        # We have row-major
        self.axisOrder = getConfigOption("imageAxisOrder")
        self.setImage(karabo_default_image())

    # ---------------------------------------------------------------------
    # Public abstract methods

    def setLookUpTable(self, lut, update=True):
        """Set a color map on the image item"""
        self.lut = lut
        stride = lut.shape[0] // 256
        self._qlut = [qRgb(*v) for v in lut[::stride, :]]
        if update:
            self.updateImage()

    def setOpts(self, update=True, **kwargs):
        if "axisOrder" in kwargs:
            value = kwargs["axisOrder"]
            if value not in ("row-major", "col-major"):
                raise ValueError("axisOrder must be either 'row-major' "
                                 "or 'col-major'")
            self.axisOrder = value
        if "lut" in kwargs:
            self.setLookUpTable(kwargs["lut"], update=update)
        if "levels" in kwargs:
            self.setLevels(kwargs["levels"], update=update)
        if update:
            self.update()

    def updateImage(self, **kwargs):
        """Update the Image"""
        defaults = {
            "autoLevels": False,
        }
        defaults.update(kwargs)
        return self.setImage(**defaults)

    def width(self):
        if self.image is None:
            return None
        axis = 0 if self.axisOrder == "col-major" else 1
        return self.image.shape[axis]

    def height(self):
        if self.image is None:
            return None
        axis = 1 if self.axisOrder == "col-major" else 0
        return self.image.shape[axis]

    def boundingRect(self):
        if self.image is None:
            return QRectF(0, 0, 0, 0)
        return QRectF(0, 0, self.width(), self.height())

    # ---------------------------------------------------------------------
    # Custom public interface

    def set_transform(self, scaling=(1, 1), translation=(0, 0), rotation=0):
        transform = QTransform()
        transform.rotate(rotation)
        transform.scale(*scaling)
        transform.translate(*translation)
        self.setTransform(transform)

        # Calculate new image item geometry
        self._origin = np.multiply(translation, scaling)
        self._rect = map_rect_to_transform(self.boundingRect(),
                                           scaling, self._origin)

    def rect(self):
        if self._rect is None:
            return self.boundingRect()
        return self._rect

    def get_color(self, x, y):
        """Get color of selected pixel. Return NULL color if there's
           no qimage or the selected pixel is out of view."""
        qimage = self._get_qimage()
        color = NULL_COLOR
        if qimage is None:
            return color

        # Map coords to qimage coords
        diff = self._slice_rect.topLeft()
        x -= int(diff.x())
        y -= int(diff.y())

        # Account downsampling (shrunk qimage)
        if self._lastDownsample is not None:
            xds, yds = self._lastDownsample
            if xds != 1:
                x = np.ceil(x / xds).astype(int)
            if yds != 1:
                y = np.ceil(y / yds).astype(int)

        # Check if calculated coord is inside the qimage
        if qimage.rect().contains(x, y):
            color = QColor(qimage.pixel(x, y))

        return color

    @Slot()
    def reset_downsampling(self, update=True):
        """Reset the downsampling of the `ImageItem`. The `update` default
        is triggered externally via a `viewbox` on resize.
        """
        self._lastDownsample = None
        if update:
            self.updateImage()

    # ---------------------------------------------------------------------
    # Events

    def mouseClickEvent(self, event):
        if (self.getViewBox().mouse_tool in self.clickTools
                and event.button() == Qt.LeftButton):
            image_pos, view_pos = self._get_mouse_positions(event.pos())
            self.clicked.emit(image_pos.x(), image_pos.y())

    def hoverEvent(self, event):
        if self.getViewBox().mouse_tool is MouseTool.Picker:
            self._hover_pointer_mode(event)

    def mouseDragEvent(self, event):
        if event.button() != Qt.LeftButton:
            event.ignore()
            return

    # ---------------------------------------------------------------------
    # Private methods

    def _hover_pointer_mode(self, event):
        if event.isExit():
            self.hovered.emit(None, None)
            return

        event.acceptDrags(Qt.LeftButton)
        event.acceptClicks(Qt.LeftButton)

        (image_pos, view_pos) = self._get_mouse_positions(event.pos())
        self.hovered.emit(image_pos.x(), image_pos.y())

    def _get_mouse_positions(self, pos):
        image_pos = self.mapToParent(*np.floor(pos))
        view_pos = image_pos - QPointF(*self._origin)
        return image_pos, view_pos

    def _get_qimage(self):
        if self.qimage is None:
            self.render()
        return self.qimage

    # ---------------------------------------------------------------------
    # Render patches

    def _set_graph_image(self, image=None, autoLevels=None, **kwargs):
        """The original setImage implementation of `pg.ImageItem`

        Note: This method is modified for lut level improvements!
        """
        new_data = False
        if image is None:
            if self.image is None:
                return
        else:
            new_data = True
            shapeChanged = (self.image is None
                            or image.shape != self.image.shape)
            image = image.view(np.ndarray)
            self.image = image
            if shapeChanged:
                self.prepareGeometryChange()
                self.informViewBoundsChanged()

        if autoLevels is None:
            if "levels" in kwargs:
                autoLevels = False
            else:
                autoLevels = True

        if autoLevels:
            image = self.image
            while image.size > 2 ** 16:
                image = image[::2, ::2]
            mn, mx = np.nanmin(image), np.nanmax(image)
            # Note: mn and mx can still be NaN if the data is all-NaN
            if mn == mx:
                # We have a valid image but same levels, but use integer
                # casting for uint8
                mnn = mn.astype(int) - 1
                # If the new min greater equal 0 take it!
                if mnn >= 0:
                    mn = mnn
                else:
                    mx += 1
            elif not np.isfinite(mn) or not np.isfinite(mx):
                # Check for `NaN` as well as `inf`
                mn = 0
                mx = 255
            kwargs["levels"] = [mn, mx]

        # Apply levels here!
        self.setOpts(update=False, **kwargs)
        self.qimage = None
        self.update()

        if new_data:
            self.sigImageChanged.emit()

    def setImage(self, image=None, **kwargs):
        """Main method to set an `image` on the `KaraboPlotImageItem`"""
        if image is None:
            return

        # Convert single-channel image to 2D array if possible
        if image.ndim == 3 and image.shape[-1] == 1:
            image = image[..., 0]

        if image.ndim == 3:
            # In case of 3 dim images we optimize out the levels calculation
            # as it is not used
            autoLevels = False
        else:
            autoLevels = self.auto_levels
        self._set_graph_image(image=image, autoLevels=autoLevels, **kwargs)

    def render(self):
        """Reimplementing for performance improvements patches"""
        # Check if image is valid
        image = self.image
        if is_image_invalid(image):
            return

        if image.ndim == 2:
            # 1. Get image according to view geometry
            image = self._slice(image)
            if is_image_invalid(image):
                return

            # 2. Downsample image according to the image geometry ratio
            #  and interpolation order
            if self.autoDownsample:
                image = self._downsample(image)
                if is_image_invalid(image):
                    return

            # 3. Clip values according to levels
            low, high = 0, 255  # default color range
            if self.levels is None:
                # Typically, levels are not `None` due to auto_levels assigning
                # the levels
                image_min, image_max = image.min(), image.max()
            else:
                # Levels are assigned and we have to clip the
                # image and rescale the minimum and maximum.
                level_min, level_max = self.levels.astype(image.dtype)
                # In place clipping with correct dtype
                image = np.clip(image, level_min, level_max, out=image)
                image_min, image_max = image.min(), image.max()
                # Calculate new color ranges with the ratio of the image
                # extrema and the preset levels.
                low, high = bytescale(np.array([image_min, image_max]),
                                      cmin=level_min, cmax=level_max,
                                      low=low, high=high)
            # 4. Rescale values to 0-255 relative to the image min/max
            # for the QImage
            image = bytescale(image, cmin=image_min, cmax=image_max,
                              low=low, high=high)

            # 6. Create QImage
            qimage = self._build_qimage(image,
                                        img_format=QImage.Format_Indexed8)

            # 7. Set color table
            if self._qlut is not None:
                qimage.setColorTable(self._qlut)
            self.qimage = qimage
        else:
            if image.shape[-1] == 3:
                # Add an opaque alpha channel to be QImage compatible
                img_format = QImage.Format_RGB888
            elif image.shape[-1] == 4:
                img_format = QImage.Format_ARGB32
            else:
                raise NotImplementedError("Formatting for image with shape "
                                          f"{image.shape} is not supported")

            # The image formats only work with uint8!
            if image.dtype != np.uint8:
                image = image.astype(np.uint8)

            self._slice_rect = QRectF(0, 0, *reversed(image.shape[:2]))
            self.qimage = self._build_qimage(image, img_format)

    def _build_qimage(self, image, img_format):
        # Transpose image array to match axis orientation.
        if self.axisOrder == "col-major":
            image = image.transpose((1, 0, 2)[:image.ndim]).copy()

        ny, nx = image.shape[:2]
        if img_format in (QImage.Format_Indexed8, QImage.Format_RGB888):
            stride = image.strides[0]
            return QImage(image, nx, ny, stride, img_format)

        qimage = QImage(image, nx, ny, img_format)
        if img_format == QImage.Format_ARGB32:
            # Swap the RGB values to BGR. Seems like Qt uses BGR color space
            qimage = qimage.rgbSwapped()
        return qimage

    def paint(self, p, *args):
        """Reimplementing for performance improvements patches"""
        if self.image is None:
            return
        if self.qimage is None:
            self.render()
            if self.qimage is None:
                return

        p.drawImage(self._slice_rect, self.qimage)

    def _slice(self, image):
        """Slice the image to be rendered wrt to the view. There's no need to
           render everything, especially if the selected view is precise and
           only shows a small part of the image. This saves us computation
           time."""

        # 0. Slice image with existing geometry
        if self._slice_rect is not None:
            x1, y1, x2, y2 = (int(c) for c in self._slice_rect.getCoords())
            return image[y1:y2, x1:x2]

        # Start calculating the view geometry
        # 1. Get image shape wrt to axis order
        shape = image.shape[:2]
        if self.axisOrder == "row-major":
            shape = shape[::-1]

        # 2. Check if coords is in the view rect.
        # If not, get new image geometry from view coords
        x1, y1 = (0, 0)
        width, height = shape

        # Get a fresh viewrect
        self._cachedView = None
        self._view_rect = self.viewRect()
        if not self._view_rect.contains(QRectF(x1, y1, width, height)):
            # Calculate image from view geometry
            x1, y1, x2, y2 = self._view_rect.getCoords()
            new_width, new_height = x2 - x1, y2 - y1
            x1 = max(int(x1 - np.ceil(new_width * 0.1)), 0)
            y1 = max(int(y1 - np.ceil(new_height * 0.1)), 0)
            x2 = min(int(x2 + np.ceil(new_width * 0.1)), width)
            y2 = min(int(y2 + np.ceil(new_height * 0.1)), height)
            width, height = x2 - x1, y2 - y1

            # Slice image according to view geometry
            image = image[y1:y2, x1:x2]

        self._slice_rect = QRectF(x1, y1, width, height)
        return image

    def _downsample(self, image):
        """Downsample the image with respect to its size and the view size
           (by calculating the pixel ratio). If the image size is larger than
           the view size, where one couldn't see the fine details, we
           downsample heavily. On the other hand, when the image size is
           smaller than the view, where one could see the pixels, we limit the
           downsampling until it's not anymore."""

        # Get downsample factor
        if self._lastDownsample is None:
            # Calculate downsample image
            self._lastDownsample = self._calculate_downsample_factor(
                x_dim=image.shape[1], y_dim=image.shape[0])

            # If still None, do nothing
            if self._lastDownsample is None:
                return
        xds, yds = self._lastDownsample

        # Scale only if downsampling of (one of) the axis is greater than 1
        if xds > 1 or yds > 1:
            scale = [1 / yds, 1 / xds]
            image = zoom(image, scale, order=self.downsample_order)

        return image

    def _calculate_downsample_factor(self, x_dim, y_dim):
        """Calculate downsampling factor based of the ratio of the screen and
           the image geometry"""

        # Reduce dimensions of image based on screen resolution
        o = self.mapToDevice(QPointF(0, 0))
        x = self.mapToDevice(QPointF(1, 0))
        y = self.mapToDevice(QPointF(0, 1))
        w = Point(x - o).length()
        h = Point(y - o).length()
        if w == 0 or h == 0:
            return

        # Calculate scaling factor for each axis.
        # - if dimension >= 1000, min scale is 2
        # - if dimension in [500, 1000), min scale is 2
        # - if dimension is smaller than 500, no scaling is done
        x_scale = np.floor(1.0 / w * 100) / 100
        y_scale = np.floor(1.0 / h * 100) / 100

        # Get minimum downsampling wrt image dimensions
        x_min_ds, y_min_ds = (1, 1)
        for dim, min_ds in DIMENSION_DOWNSAMPLE:
            if x_dim > dim:
                x_min_ds = min_ds
            if y_dim > dim:
                y_min_ds = min_ds

        # Correct downsampling wrt if less than prescribed minimum
        xds = x_scale if x_scale > x_min_ds else x_min_ds
        yds = y_scale if y_scale > y_min_ds else y_min_ds

        return xds, yds

    def setLevels(self, levels, update=True):
        if levels is not None:
            levels = np.asarray(levels)
        if not fn.eq(levels, self.levels):
            self.levels = levels
            if update:
                self.updateImage()

    def viewTransformChanged(self):
        """Reimplemented to control image downsampling/slicing to view.

           When the imageItem is zoomed, we recalculate the image slice for the
           new view ranges and the downsampling of the image. If panned, we
           only consider calculating the image slice by checking if the view
           exceeds the current slice rect."""
        super().viewTransformChanged()

        self.qimage = None
        zoomed = True

        # Note: Invalidate the cachedView before getting a new `viewRect`
        view_rect = self.viewRect()
        # Check if image is zoomed. Checking with width is already sufficient.
        if self._view_rect is not None:
            diff = abs(view_rect.width() - self._view_rect.width())
            zoomed = diff > 0.01 * view_rect.width()
        self._view_rect = view_rect

        if zoomed:
            self._lastDownsample = None
            self._slice_rect = None
        elif self._slice_rect is not None:
            # Image is just panned.
            # Check if there is a need to get a new slice of QImage
            view_x1, view_y1, view_x2, view_y2 = \
                view_rect.getCoords()
            slice_x1, slice_y1, slice_x2, slice_y2 = \
                self._slice_rect.getCoords()

            # Get current image dimensions
            # (via boundingRect to consider axis major)
            im_x1, im_y1, im_x2, im_y2 = self.boundingRect().getCoords()

            # Get to-be-rendered image dimensions (wrt view)
            im_x1 = max(view_x1, im_x1)
            im_y1 = max(view_y1, im_y1)
            im_x2 = min(view_x2, im_x2)
            im_y2 = min(view_y2, im_y2)

            # Check if new required image is out of bounds of the slice
            x_out = im_x1 < slice_x1 or im_y1 < slice_y1
            y_out = im_x2 > slice_x2 or im_y2 > slice_y2
            out_of_bounds = x_out or y_out

            # Reset slice if out of bounds,
            # Else do nothing, sliced QImage is still sufficient for the view
            if out_of_bounds:
                self._slice_rect = None
            else:
                return

        self.updateImage()

    def informViewBoundsChanged(self):
        super().informViewBoundsChanged()
        self.reset_downsampling(update=False)
        self._slice_rect = None

    def clear(self):
        self.image = None
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.update()


def is_image_invalid(image):
    """Convenience function to check if we have a valid `image` with a size"""
    return image is None or image.size == 0



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Configurable, Hash, String, VectorHash
from karabogui.configurator.api import TableDialog
from karabogui.testing import (
    click_button, get_class_property_proxy, set_proxy_hash)


class TableSchema(Configurable):
    foo = String(
        defaultValue="NoString")
    bar = String(
        displayType="State",
        displayedName="Bar",
        defaultValue="ON")


class BigTableSchema(Configurable):
    a = String()
    b = String()
    c = String()
    d = String()
    e = String()


TABLE_HASH = Hash("prop", [Hash("foo", "1", "bar", "hello"),
                           Hash("foo", "2", "bar", "test"),
                           Hash("foo", "3", "bar", "No"),
                           Hash("foo", "4", "bar", "Jo")])


class Object(Configurable):
    prop = VectorHash(rows=TableSchema)
    bigProp = VectorHash(rows=BigTableSchema)


def test_table_view_dialog(gui_app, subtests):
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    dialog = TableDialog(proxy, True)
    assert dialog is not None
    assert dialog.width() == 450
    set_proxy_hash(proxy, TABLE_HASH)

    assert dialog.width() == 450
    # Finish dialog, destroy widget
    dialog.done(1)
    assert dialog.controller is not None
    assert dialog.controller.widget is None

    # New dialog, with big table schema
    proxy = get_class_property_proxy(Object.getClassSchema(), "bigProp")
    dialog = TableDialog(proxy, True)
    assert dialog is not None
    assert dialog.width() == 600
    # Finish dialog without success
    dialog.done(0)
    assert dialog.controller is not None
    assert dialog.controller.widget is None
    assert dialog.toolbar is None

    # New dialog, launch actions
    with subtests.test("Test the table view action buttons"):
        proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
        dialog = TableDialog(proxy, True)
        assert dialog is not None
        assert dialog.width() == 450
        set_proxy_hash(proxy, TABLE_HASH)
        assert dialog.controller.sourceModel() is not None

        data = dialog.controller.sourceModel().index(0, 0).data()
        assert data == "1"
        dialog.controller.tableWidget().selectRow(0)
        click_button(dialog.toolbar._move_down_button)
        data = dialog.controller.sourceModel().index(0, 0).data()
        assert data == "2"
        dialog.controller.tableWidget().selectRow(1)
        click_button(dialog.toolbar._move_up_button)
        data = dialog.controller.sourceModel().index(0, 0).data()
        assert data == "1"

        click_button(dialog.toolbar._remove_button)
        data = dialog.controller.sourceModel().index(0, 0).data()
        assert data == "2"

        assert dialog.controller.sourceModel().rowCount() == 3
        dialog.controller.tableWidget().selectRow(2)
        data = dialog.controller.sourceModel().index(2, 0).data()
        assert data == "4"
        # Adding adds a default value
        click_button(dialog.toolbar._add_button)
        assert dialog.controller.sourceModel().rowCount() == 4
        data = dialog.controller.sourceModel().index(3, 0).data()
        assert data == "NoString"

        # Duplicate copies
        dialog.controller.tableWidget().selectRow(2)
        click_button(dialog.toolbar._du_button)
        assert dialog.controller.sourceModel().rowCount() == 5
        data = dialog.controller.sourceModel().index(3, 0).data()
        assert data == "4"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import FloatSpinBoxModel
from karabo.native import Configurable, Double, Unit
from karabogui.binding.api import build_binding
from karabogui.const import IS_MAC_SYSTEM
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..floatspinbox import FloatSpinBox


class Object(Configurable):
    prop = Double(minInc=-10.0, maxInc=10.0, absoluteError=0.5,
                  unitSymbol=Unit.METER)


class Other(Configurable):
    prop = Double(minInc=1.0, maxInc=42.0)


@pytest.fixture
def floatspinbox_setup(gui_app):
    # setup
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    model = FloatSpinBoxModel(step=0.1, decimals=5)
    controller = FloatSpinBox(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None
    controller.set_read_only(False)
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_floatspinbox_basics(floatspinbox_setup):
    controller, proxy = floatspinbox_setup
    # focus policy
    assert controller.widget.focusPolicy() == Qt.StrongFocus

    # set value
    set_proxy_value(proxy, "prop", 5.0)
    assert controller.widget.value() == 5.0

    # edit value
    controller.widget.setValue(3.0)
    assert proxy.edit_value == 3.0


def test_floatspinbox_schema_update(gui_app):
    proxy = get_class_property_proxy(Other.getClassSchema(), "prop")
    controller = FloatSpinBox(proxy=proxy, model=FloatSpinBoxModel())
    controller.create(None)
    assert controller.widget.suffix() == ""

    assert controller.widget.minimum() == 1.0
    assert controller.widget.maximum() == 42.0

    build_binding(Object.getClassSchema(),
                  existing=proxy.root_proxy.binding)

    assert controller.widget.minimum() == -10.0
    assert controller.widget.maximum() == 10.0
    assert controller.widget.suffix() == " m"


def test_floatspinbox_actions(floatspinbox_setup, mocker):
    controller, _ = floatspinbox_setup
    actions = controller.widget.actions()
    change_step, change_decimals, formatting = actions

    assert "step" in change_step.text().lower()
    assert "decimals" in change_decimals.text().lower()

    sym = "karabogui.controllers.edit.floatspinbox.QInputDialog"
    QInputDialog = mocker.patch(sym)
    QInputDialog.getDouble.return_value = 0.25, True
    QInputDialog.getInt.return_value = 9, True

    change_step.trigger()
    assert controller.model.step == 0.25

    change_decimals.trigger()
    assert controller.model.decimals == 9


@pytest.mark.skipif(IS_MAC_SYSTEM, reason="Fonts are different on MacOS")
def test_floatspinbox_font_setting(floatspinbox_setup, mocker):
    controller, _ = floatspinbox_setup
    settings = "{ font: normal; font-size: 10pt; }"
    assert settings in controller.widget.styleSheet()

    path = "karabogui.controllers.edit.floatspinbox.FormatLabelDialog"
    dialog = mocker.patch(path)
    dialog().exec.return_value = QDialog.Accepted
    dialog().font_size = 11
    dialog().font_weight = "bold"
    controller._format_field()

    settings = "{ font: bold; font-size: 11pt; }"
    assert settings in controller.widget.styleSheet()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtWidgets import QDialog

from karabo.common.api import WeakMethodRef
from karabogui import messagebox
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.request import call_device_slot
from karabogui.util import get_spin_widget
from karabogui.widgets.log import LogWidget

from .utils import get_dialog_ui


class LogDialog(QDialog):
    def __init__(self, server_id, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("log_dialog.ui"), self)
        flags = Qt.WindowCloseButtonHint | Qt.WindowTitleHint
        self.setWindowFlags(self.windowFlags() | flags)

        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        self.setSizeGripEnabled(True)
        self.setWindowTitle(f"Server log for serverId: {server_id}")
        self.log_widget = LogWidget(resize_contents=True, parent=self)
        self.log_layout.insertWidget(0, self.log_widget)

        self.server_id = server_id
        self.ui_server_id.setText(server_id)
        self.spin_widget = get_spin_widget(icon="wait-black",
                                           scaled_size=QSize(16, 16))
        self.spin_widget.setVisible(False)
        self.bottom_layout.insertWidget(0, self.spin_widget)
        self.ui_request.clicked.connect(self.request_logger_data)
        self.ui_fetching_data.setVisible(False)
        self.request_logger_data()
        self.event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network
        }
        register_for_broadcasts(self.event_map)

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    def done(self, result):
        """Stop listening for broadcast events"""
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    def request_handler(self, success, reply):
        self.spin_widget.setVisible(False)
        self.ui_fetching_data.setVisible(False)
        if not success:
            messagebox.show_error(
                f"Could not fetch the logs of server <b>{self.server_id}</b>.",
                parent=self)
            return

        data = reply["content"]
        self.log_widget.initialize(data)

    def keyPressEvent(self, event):
        if (event.key() in (Qt.Key_Enter, Qt.Key_Return) and
                self.focusWidget() == self.ui_number_logs):
            self.request_logger_data()
            event.accept()
            return
        return super().keyPressEvent(event)

    @Slot()
    def request_logger_data(self):
        self.spin_widget.setVisible(True)
        self.ui_fetching_data.setVisible(True)
        call_device_slot(WeakMethodRef(self.request_handler), self.server_id,
                         "slotLoggerContent",
                         logs=self.ui_number_logs.value())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.graph.plots.api import RangeDialog


def test_range_dialog(gui_app):
    """Test the range dialog"""
    config = {
        "x_min": 0,
        "x_max": 20,
        "x_autorange": False,
    }
    actual = {
        "x_min": 2,
        "x_max": 3,
        "x_autorange": False,
    }
    dialog = RangeDialog(config, actual)

    assert dialog.ui_view_min.text() == "2.00e+00"
    assert dialog.ui_view_max.text() == "3.00e+00"

    assert dialog.ui_min.text() == "0"
    assert dialog.ui_max.text() == "20"
    assert dialog.ui_min.isEnabled()
    assert dialog.ui_max.isEnabled()
    assert not dialog.ui_autorange.isChecked()

    dialog.ui_autorange.setChecked(True)
    assert not dialog.ui_min.isEnabled()
    assert not dialog.ui_max.isEnabled()

    new = {
        "x_min": 0,
        "x_max": 20,
        "x_autorange": True}

    assert dialog.limits == new



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QEventLoop, QPoint, QSize, Qt, QTimer, Slot
from qtpy.QtWidgets import QApplication, QDialog, QMenu, QMessageBox, QStyle

from .utils import get_dialog_ui

MESSAGE_POPUP = 60  # Seconds
MESSAGEBOX_STANDARD_PIXMAPS = {
    QMessageBox.Information: QStyle.SP_MessageBoxInformation,
    QMessageBox.Warning: QStyle.SP_MessageBoxWarning,
    QMessageBox.Critical: QStyle.SP_MessageBoxCritical,
}


class KaraboMessageBox(QDialog):
    """A pop up message box which will automatically close after
    `MESSAGE_POPUP` seconds!
    """
    existing = set()

    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("messagebox.ui"), self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        # Set focus on the OK button
        self.ok_button.clicked.connect(self.accept)
        self.ok_button.setFocus()

        # Initialize the "Show details" part of the widget
        self.show_details_button.clicked.connect(self._show_details)
        self.show_details_button.setVisible(False)
        self.details_widget.setVisible(False)
        self.details_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.details_widget.customContextMenuRequested.connect(
            self._show_context_menu)
        self._is_showing_details = False
        self._size_hint = QSize()

        # Set up a ticker to eventually close the pop if not already closed!
        self._ticker = QTimer(self)
        self._ticker.setInterval(MESSAGE_POPUP * 1000)
        self._ticker.setSingleShot(True)
        self._ticker.timeout.connect(self.close)
        # We already start ticking here!
        self._ticker.start()
        self.existing.add(self)

    # ----------------------------------------------------------------------
    # Qt methods

    def closeEvent(self, event):
        """Close the ticker before closing the dialog"""
        if self._ticker.isActive():
            self._ticker.stop()
        # Graceful unregister with discard!
        self.existing.discard(self)
        super().closeEvent(event)
        flags = QEventLoop.AllEvents
        QApplication.processEvents(flags, 100)

    def sizeHint(self):
        return self._size_hint

    # ----------------------------------------------------------------------
    # Reimplemented QMessagebox methods

    def setIcon(self, icon):
        style = QApplication.style()
        icon = style.standardIcon(MESSAGEBOX_STANDARD_PIXMAPS[icon])
        icon_size = style.pixelMetric(QStyle.PM_MessageBoxIconSize)
        self.icon_label.setPixmap(icon.pixmap(icon_size, icon_size))

    def setText(self, text):
        """Adjust the dialog size after setting the text to recompute for an
           ideal size."""
        self.text_label.setText(text)
        self.adjustSize()

    def setDetailedText(self, details):
        """Adjust the dialog size after setting the text to recompute for an
           ideal size."""
        self.show_details_button.setVisible(True)
        self.details_textedit.setText(details)
        self.adjustSize()

    # ----------------------------------------------------------------------
    # Private methods

    @Slot()
    def _show_details(self):
        """Show/hide the details widget area and modify the button text when
           the "Show(Hide) Details..." button is clicked. This also recomputes
           for the ideal size of the dialog."""
        button_text = {False: "Show Details...", True: "Hide Details..."}
        self._is_showing_details = not self._is_showing_details
        self._size_hint = self.size()

        self.show_details_button.setText(button_text[self._is_showing_details])
        self.details_widget.setVisible(self._is_showing_details)
        self.adjustSize()

    @Slot(QPoint)
    def _show_context_menu(self, pos):
        """Show a context menu"""
        menu = QMenu(self)
        widget = self.details_textedit
        select_action = menu.addAction('Select All')
        select_action.triggered.connect(widget.selectAll)
        enable_select = len(widget.toPlainText()) > 0
        select_action.setEnabled(enable_select)

        copy_action = menu.addAction('Copy Selected')
        copy_action.triggered.connect(widget.copy)
        enable_cp = not widget.textCursor().selection().isEmpty()
        copy_action.setEnabled(enable_cp)
        menu.exec(widget.viewport().mapToGlobal(pos))



from itertools import cycle
from weakref import WeakValueDictionary

from qtpy.QtGui import QColor
from qtpy.QtWidgets import QAction, QDialog
from traits.api import Bool, Instance

from karabo.common.scenemodel.api import (
    KARABO_CURVE_OPTIONS, CurveType, build_graph_config,
    extract_graph_curve_option, restore_graph_config)
from karabogui import icons
from karabogui.controllers.arrays import get_array_data
from karabogui.controllers.base import BaseBindingController
from karabogui.graph.common.colors import get_pen_cycler
from karabogui.graph.common.toolbar.widgets import create_button
from karabogui.graph.plots.base import KaraboPlotView
from karabogui.graph.plots.dialogs import CurveOptionsDialog
from karabogui.graph.plots.dialogs.data_analysis import DataAnalysisDialog
from karabogui.graph.plots.dialogs.transform_config import TransformDialog
from karabogui.graph.plots.utils import (
    create_curve_options, generate_baseline, generate_down_sample,
    get_view_range)


class BaseArrayGraph(BaseBindingController):
    _curves = Instance(WeakValueDictionary, args=())
    _pens = Instance(cycle, allow_none=False)

    _has_curve_options = Bool(False)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.stateChanged.connect(self._change_model)
        widget.add_legend(visible=False)
        widget.add_cross_target()
        widget.add_roi()
        toolbar = widget.add_toolbar()
        config = build_graph_config(self.model)
        widget.restore(config)
        widget.enable_data_toggle()
        widget.enable_export()

        trans_action = QAction("X-Transformation", widget)
        trans_action.triggered.connect(self.configure_transformation)

        widget.addAction(trans_action)

        if KARABO_CURVE_OPTIONS in self.model.copyable_trait_names():
            self._has_curve_options = True
            curve_options = QAction("Curve Options", widget)
            curve_options.setIcon(icons.settings)
            curve_options.triggered.connect(self.show_options_dialog)
            widget.addAction(curve_options)

        # Add first curve for the main proxy
        self._add_curve(self.proxy, widget=widget)

        button = create_button(
            icon=icons.data_analysis, checkable=False, tooltip="Data Analysis",
            on_clicked=self.show_data_analysis_dialog)
        toolbar.add_button(button)
        return widget

    def show_data_analysis_dialog(self):
        config = build_graph_config(self.model)
        data_analysis_dialog = DataAnalysisDialog(
            proxies=self.proxies, config=config, parent=self.widget)
        data_analysis_dialog.show()

    def __pens_default(self):
        return get_pen_cycler()

    # ----------------------------------------------------------------
    # Plot Options

    def show_options_dialog(self):
        config = build_graph_config(self.model)
        options = config.get(KARABO_CURVE_OPTIONS, {})
        curve_options = create_curve_options(
            curves=self._curves, options=options,
            curve_type=CurveType.Curve)

        dialog = CurveOptionsDialog(curve_options, parent=self.widget)
        if dialog.exec() != QDialog.Accepted:
            return

        curve_options = dialog.get_curve_options()
        if curve_options is None:
            self.reset_curve_options()
            return

        if curve_options:
            self.set_curve_options(curve_options)

    def set_curve_options(self, curve_options: dict):
        """Method to set the curve options on the model and plot"""
        self._change_model(
            {KARABO_CURVE_OPTIONS: list(curve_options.values())})
        self.widget.apply_curve_options(curve_options)

    def reset_curve_options(self):
        """Restore the default curve options."""
        self.reset_traits(["_pens"])
        options = {}
        for proxy, curve in self._curves.items():
            pen = next(self._pens)
            options[curve] = {
                "name": proxy.key,
                "curve_type": CurveType.Curve,
                "pen_color": pen.color().name()}

        self.model.curve_options = []
        self.widget.apply_curve_options(options)

    # ----------------------------------------------------------------

    def add_proxy(self, proxy):
        """Add a proxy curve to the plot widget

        We only rely here on the `key` information of the binding, which is
        always available. This is required for the plot item `name`.
        We are protected by the base binding controller that a proxy is not
        added multiple times.
        """
        self._add_curve(proxy)
        if len(self._curves) > 1:
            self.widget.set_legend(True)
        return True

    def remove_proxy(self, proxy):
        """Remove a `proxy` from the controller"""
        if self._has_curve_options:
            options = [option for option in self.model.curve_options
                       if option.key != proxy.key]
            self.model.trait_set(curve_options=options)

        item = self._curves.pop(proxy)
        self.widget.remove_item(item)
        item.deleteLater()
        legend_visible = len(self._curves) > 1
        self.widget.set_legend(legend_visible)
        return True

    def _add_curve(self, proxy, widget=None):
        """The widget is passed as an argument in create_widget as it is not
           yet bound to self.widget then"""
        if widget is None:
            widget = self.widget

        name = proxy.key
        pen = next(self._pens)
        if self._has_curve_options:
            options = extract_graph_curve_option(self.model, name)
            if options:
                pen_color = options.get("pen_color")
                pen.setColor(QColor(pen_color))
                name = options.get("name")

        curve = widget.add_curve_item(name=name, pen=pen)
        self._curves[proxy] = curve

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**restore_graph_config(content))

    def configure_transformation(self):
        content, ok = TransformDialog.get(build_graph_config(self.model),
                                          parent=self.widget)
        if ok:
            self.model.trait_set(**content)
            for proxy in self.proxies:
                self.value_update(proxy)

    def value_update(self, proxy):
        if proxy not in self._curves:
            return

        plot = self._curves[proxy]
        # NOTE: With empty data we will clear the plot!
        y, _ = get_array_data(proxy, default=[])
        if not len(y):
            plot.setData([], [])
            return

        # NOTE: WE cast boolean as int, as numpy method is deprecated
        if y.dtype == bool:
            y = y.astype(int)

        model = self.model
        # Generate the baseline for the x-axis
        x = generate_baseline(y, offset=model.offset, step=model.step)

        rect = get_view_range(plot)
        x, y = generate_down_sample(y, x=x, rect=rect, deviation=True)
        plot.setData(x, y)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from traits.api import BaseCBool, BaseFloat, BaseRange


class NumpyRange(BaseRange):
    """A fast-validating trait type who casts to a `numpy` dtype on validate
    """

    def __init__(self, dtype=None, **kwargs):
        self._dtype = dtype
        super().__init__(**kwargs)

    def validate(self, object, name, value):
        """ Validate that the value is in the specified range.
        """
        if type(value) is str:
            try:
                return self._dtype(value)
            except (OverflowError, ValueError):
                self.error(object, name, value)

        return self._dtype(super().validate(object, name, value))


class Float(BaseFloat):
    """ A fast-validating trait type who can handle numpy dtypes and coerce
    """

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait."""
        value_type = type(value)
        if value_type in (float, np.float32, np.float64):
            return value
        else:
            try:
                return float(value)
            except Exception:
                self.error(object, name, value)


class CBool(BaseCBool):
    """ A fast-validating trait type who can coerce, but does not accept `None`
    """

    def validate(self, object, name, value):
        """Reimplemented validate for `None` strings

        Note: `None` check can be removed when the configs are sanitized
        """
        if value in ("None", None):
            self.error(object, name, value)
        try:
            return bool(value)
        except Exception:
            self.error(object, name, value)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 3, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path
import shutil
import webbrowser
from enum import Enum
from pathlib import Path

from qtpy.QtCore import QPoint, QSize, Qt, Slot
from qtpy.QtGui import QColor, QIcon
from qtpy.QtWidgets import (
    QAction, QActionGroup, QFrame, QLabel, QMainWindow, QMenu, QMessageBox,
    QPushButton, QSizePolicy, QSplitter, QTextBrowser, QToolButton, qApp)

import karabogui.access as krb_access
from karabo.common.api import KARABO_PROJECT_MANAGER
from karabo.common.project.api import get_project_models
from karabo.common.project.utils import check_instance_duplicates
from karabo.native import AccessLevel
from karabogui import const, icons, messagebox
from karabogui.access import ACCESS_LEVELS, AccessRole
from karabogui.background import background
from karabogui.dialogs.api import (
    AboutDialog, ApplicationConfigurationDialog, ClientTopologyDialog,
    DataViewDialog, DevelopmentTopologyDialog, GuiSessionInfo,
    ProjectTopologyDialog, UpdateDialog, UserSessionDialog)
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts)
from karabogui.indicators import get_processing_color
from karabogui.logger import StatusLogWidget, get_logger
from karabogui.panels.api import (
    ConfigurationPanel, DevicePanel, PanelContainer, ProjectPanel,
    TopologyPanel)
from karabogui.programs.register_protocol import register_protocol
from karabogui.programs.utils import (
    create_linux_desktop_file, run_concert, save_concert_file)
from karabogui.project.restore import get_restore_data
from karabogui.request import call_device_slot
from karabogui.singletons.api import (
    get_config, get_db_conn, get_network, get_project_model)
from karabogui.util import (
    convert_npy_to_csv, generateObjectName, get_reason_parts, getOpenFileName,
    getSaveFileName, move_to_cursor, process_qt_events)
from karabogui.wizards import TipsTricksWizard

SERVER_INFO_WIDTH = 250

CONFIGURATOR_TITLE = 'Configuration Editor'
SYSTEM_TOPOLOGY_TITLE = 'System Topology'
DEVICE_TOPOLOGY_TITLE = 'Device Topology'
PROJECT_TITLE = 'Projects'

_PANEL_TITLE_CONFIG = {
}

MENU_HEIGHT = 40

ACCESS_LEVEL_INFO = """<html><head/><body><p><span style=" font-size:10pt;
font-style:italic; font-weight:normal; color:#242424;">Currently logged in as
temporary user <b>'{user}'</b> with the access level <b>'{access_level}'</b>.
Do you really want to end the Temporary session?</span></p>
"""


class MainWindowBanner(QTextBrowser):
    """The MainWindow Banner"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)
        self.setFixedHeight(MENU_HEIGHT)
        self.setFrameStyle(QFrame.NoFrame)
        self.setSizePolicy(QSizePolicy.Expanding,
                           QSizePolicy.Fixed)
        self.setAlignment(Qt.AlignCenter)
        self.setTextInteractionFlags(
            Qt.LinksAccessibleByMouse | Qt.TextSelectableByMouse)
        self.setOpenExternalLinks(True)

        self.setObjectName(generateObjectName(self))
        self.setFocusPolicy(Qt.NoFocus)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            self._show_context_menu)
        self.setLineWidth(1)
        self.set_banner("")

    def set_banner(self, text="", foreground="black",
                   background="transparent"):
        """Set the banner of the main window"""
        if text:
            text = f'<p align="center">{text}</p>'
            style = QFrame.Box
        else:
            background = "transparent"
            style = QFrame.NoFrame

        self.setStyleSheet(f"QAbstractScrollArea#{self.objectName()} "
                           f"{{ color: {foreground};"
                           f"background-color: {background}; }}")
        # text
        self.setFrameStyle(style)
        self.setHtml(text)

    @Slot(QPoint)
    def _show_context_menu(self, pos):
        """Show a context menu"""
        menu = QMenu(self)
        select_action = menu.addAction("Select All")
        select_action.triggered.connect(self.selectAll)
        enable_select = len(self.toPlainText()) > 0
        select_action.setEnabled(enable_select)

        copy_action = menu.addAction("Copy Selected")
        copy_action.triggered.connect(self.copy)
        enable_copy = not self.textCursor().selection().isEmpty()
        copy_action.setEnabled(enable_copy)
        menu.exec(self.viewport().mapToGlobal(pos))


class PanelAreaEnum(Enum):
    """The different panel areas in the main window
    """
    Left = 0
    Middle = 1
    Right = 2


_PANELS = {
    # Title: (class, position)
    CONFIGURATOR_TITLE: (ConfigurationPanel, PanelAreaEnum.Right),
    SYSTEM_TOPOLOGY_TITLE: (TopologyPanel, PanelAreaEnum.Left),
    DEVICE_TOPOLOGY_TITLE: (DevicePanel, PanelAreaEnum.Left),
    PROJECT_TITLE: (ProjectPanel, PanelAreaEnum.Left),
}

SETTINGS_TITLE = '&Application Settings'
GEOMETRY_TITLE = "&Window Geometry"
GRAFANA_LINK = "https://ctrend.xfel.eu/"
PAGES_FW_LINK = "https://karabo.pages.xfel.eu/Framework/"

PAGES_LINK = "https://karabodevices.pages.xfel.eu"
PAGES_MACRO_LINK = f"{PAGES_LINK}/howtomacro/"
PAGES_GUI_LINK = f"{PAGES_LINK}/howtogui/"
KEYBOARD_SHORTCUTS = f"{PAGES_LINK}/howtogui/keyboard_shortcuts.html"


class MainWindow(QMainWindow):
    """The main window of the application which includes all relevant panels
    and the main toolbar.
    """

    def __init__(self):
        super().__init__()

        # Create projects folder, if not exists
        if not os.path.exists(const.KARABO_PROJECT_FOLDER):
            os.makedirs(const.KARABO_PROJECT_FOLDER, exist_ok=True)

        # reference to view menu and its submenus {menuName: QMenu}
        self.viewMenus = {}

        self._setupActions()
        self._setupMenuBar()
        self._setupToolBar()
        self._setupStatusBar()

        # Keep track of the panels!
        self._active_panels = {}
        self._panel_areas = {}
        self._setupPanelAreas()
        for name in _PANELS:
            self._open_panel(name)

        self.title_info = {
            "project": None,
            "version": const.GUI_VERSION_LONG,
            "topic": None
        }
        self._set_window_title()
        # Connect to some important network signals
        network = get_network()
        network.signalServerConnectionChanged.connect(
            self.onServerConnectionChanged)
        network.signalNetworkPerformance.connect(
            self.onNetworkPerformance)

        # Register to KaraboBroadcastEvent, Note: unregister_from_broadcasts is
        # not necessary
        event_map = {
            KaraboEvent.ServerInformationUpdate: self._event_server_connection,
            KaraboEvent.BigDataProcessing: self._event_big_data,
            KaraboEvent.DatabaseIsBusy: self._event_db_processing,
            KaraboEvent.MaximizePanel: self._event_container_maximized,
            KaraboEvent.MinimizePanel: self._event_container_minimized,
            KaraboEvent.LoginUserChanged: self._event_access_level,
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.ProjectUpdated: self._event_project_updated,
            KaraboEvent.ServerNotification: self._event_server_notification,
            KaraboEvent.ProjectName: self._event_project_title,
        }
        register_for_broadcasts(event_map)

        self.readSettings()

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_server_connection(self, data):
        self.update_server_connection(data)

    def _event_network(self, data):
        status = data.get('status', False)
        if not status:
            # On disconnect, we select again the system topology!
            container = self._panel_areas[PanelAreaEnum.Left]
            tab = self._active_panels[SYSTEM_TOPOLOGY_TITLE]
            container.setCurrentIndex(container.indexOf(tab))

            # Set a default title on disconnect
            self._set_window_title(project=None, topic=None)

    def _event_big_data(self, data):
        """Show the big data latency including value set in the gui"""
        name = data['name']
        proc = data['proc']
        self.ui_big_data.setText(f"{name} - {proc:.3f} s")

    def _event_db_processing(self, data):
        """This method gets called whenever the database is switching its
        processing mode.

        Scenes, Macros, and Configurations should not be editable while the
        project DB is saving data.
        """
        enable = not data.get('is_processing')
        # Update toolbar
        self._enable_toolbar(enable)

        # Update all open scenes and macros
        container = self._panel_areas[PanelAreaEnum.Middle]
        for panel in container.panel_set:
            panel.setEnabled(enable)

        # Update configuration panels
        container = self._panel_areas[PanelAreaEnum.Right]
        for panel in container.panel_set:
            panel.setEnabled(enable)

    def _event_container_maximized(self, data):
        """The given `panel_container` is about to be maximized.
        """
        panel_container = data['container']
        for area_enum, area_container in self._panel_areas.items():
            if area_container is not panel_container:
                area_container.minimize(True)

    def _event_container_minimized(self, data):
        """The given `panel_container` is about to be minimized.
        """
        panel_container = data['container']
        for area_enum, area_container in self._panel_areas.items():
            if area_container is not panel_container:
                area_container.minimize(False)

    def _event_server_notification(self, data):
        text = data.get('message', '')
        foreground = QColor(data.get('foreground', 'black')).name()
        background = QColor(data.get('background', 'white')).name()
        self.notification_banner.set_banner(text, foreground, background)

    def _event_access_level(self, data):
        self.onUpdateAccessLevel()
        self._set_window_title()

    def _event_project_updated(self, data):
        """Projects were externally updated and we check if we are affected"""
        project = get_project_model().root_model
        if project is None:
            return

        projects = get_project_models(project)
        external_uuids = data['uuids']
        conflict = [item for item in projects if item.uuid in external_uuids]
        for model in conflict:
            model.conflict = True
            get_logger().warning(f"The project {model.simple_name} was "
                                 f"updated externally. Please reload the "
                                 f"project to synchronize the changes.")

    # -----------------------------------------------------------------------

    def _event_project_title(self, data):
        name = data["simple_name"]
        self._set_window_title(project=name)

    def _set_window_title(self, **kwargs):
        self.title_info.update(kwargs)
        titles = ["{}", "European XFEL - Karabo GUI {}", "Topic: {}"]
        title = " - ".join([info.format(value) for info, value in
                            zip(titles, self.title_info.values())
                            if value is not None])
        if get_config()["username"] and krb_access.is_authenticated():
            title = f"{title} - User: {get_config()['username']}"
        if krb_access.TEMPORARY_SESSION_USER is not None:
            title = (
                f"{title} - Temporary Session User: "
                f"{krb_access.TEMPORARY_SESSION_USER}")
        self.setWindowTitle(title)

    def readSettings(self):
        window_geometry = get_config()["main_geometry"]
        if window_geometry:
            self.restoreGeometry(window_geometry)
        else:
            self.adjustSize()

    def update_server_connection(self, data=None):
        """Update the status bar with our broker connection information
        """
        if data is not None:
            topic = data["topic"]
            # Store this information in the config singleton!
            get_config()["broker_topic"] = topic
            self._set_window_title(topic=topic)

            hostname = data["hostname"]
            hostport = data["hostport"]
            info = f"<b>{hostname}:{hostport} ({topic})</b>"
            self.serverInfo.setText(info)

            # Place some user log info messages
            logger = get_logger()
            text = f"Successfully connected to gui server (topic): {info}"
            logger.info(text)
            # If the server is read only, we notify as well
            read_only = data.get("readOnly", False)
            if read_only:
                text = "The gui server device is configured as <b>readOnly</b>"
                logger.info(text)
        else:
            self.serverInfo.setText("")
            self.tbUserSession.setVisible(False)

    # --------------------------------------
    # Qt virtual methods

    def sizeHint(self):
        return QSize(1200, 800)

    def closeEvent(self, event):
        if not self._quit():
            event.ignore()
            return

        event.accept()
        QMainWindow.closeEvent(self, event)
        # Process eventual events to gracefully close main window
        process_qt_events(timeout=1000)
        qApp.quit()

    # --------------------------------------
    # public methods

    def addPanel(self, panel, area_enum):
        """Add a panel to an area's container in this window
        """
        container = self._panel_areas[area_enum]
        container.addPanel(panel)

    def removePanel(self, panel, area_enum):
        """Remove a panel from a container in this window
        """
        container = self._panel_areas[area_enum]
        container.removePanel(panel)

    def selectPanel(self, panel, area_enum):
        """This method raises a panel to the top of the focus stack, regardless
        of its current docked/undocked status.
        """
        container = self._panel_areas[area_enum]
        index = container.indexOf(panel)
        if index > -1:
            container.setCurrentIndex(index)
        else:
            panel.activateWindow()
            panel.raise_()

    def dockPanel(self, name):
        """Public method to ensure that a panel by `name` is docked"""
        if name in self._active_panels:
            panel = self._active_panels[name]
            if not panel.is_docked:
                panel.onDock()

    def onUpdateAccessLevel(self):
        global_access_level = krb_access.GLOBAL_ACCESS_LEVEL
        highest_access_level = global_access_level
        if krb_access.is_authenticated():
            highest_access_level = krb_access.HIGHEST_ACCESS_LEVEL
        # Build the access level menu
        self.mAccessLevel.clear()
        for level in ACCESS_LEVELS.values():
            if highest_access_level >= level:
                action = self.access_level_actions[level]
                self.mAccessLevel.addAction(action)

        # Show a check next to the current access level
        for action in self.access_level_actions.values():
            action.setChecked(False)

        checked_action = self.access_level_actions.get(global_access_level)
        if checked_action is not None:
            checked_action.setChecked(True)
        # SERVER_READ_ONLY can be None for non-authenticated gui server.
        allow_user_session = krb_access.SERVER_READ_ONLY is False
        self.tbUserSession.setVisible(allow_user_session)

    def setSessionButton(self, icon: QIcon, tooltip: str) -> None:
        """Set the icon and tooltip for the Temporary Session button"""
        self.tbUserSession.setToolTip(tooltip)
        self.tbUserSession.setIcon(icon)

    def enableGuiExtensionUpdate(self):
        """This method is called on gui launch if a new version of GUI
        Extensions is available.
        """
        self.update_extensions_action.setVisible(True)

    # --------------------------------------
    # private methods

    def _setupActions(self):
        text = "Change access level"
        self.tbAccessLevel = QToolButton(self)
        self.tbAccessLevel.setIcon(icons.lock)
        self.tbAccessLevel.setToolTip(text)
        self.tbAccessLevel.setStatusTip(text)
        self.tbAccessLevel.setPopupMode(QToolButton.InstantPopup)
        self.tbAccessLevel.setEnabled(False)

        text = "User Session Info"
        self.tbUserSession = QAction(icons.switchNormal, f"&{text}", self)
        self.tbUserSession.setToolTip(text)
        self.tbUserSession.setStatusTip(text)
        self.tbUserSession.triggered.connect(self.onTemporarySession)

        self.agAccessLevel = QActionGroup(self)
        self.agAccessLevel.triggered.connect(self.onChangeAccessLevel)

        self.access_level_actions = {}
        for name, level in ACCESS_LEVELS.items():
            action = QAction(name, self)
            action.setStatusTip(text)
            action.setToolTip(text)
            action.setCheckable(True)
            action.setData(level)
            self.agAccessLevel.addAction(action)
            self.access_level_actions[level] = action

        self.mAccessLevel = QMenu()
        self.onUpdateAccessLevel()
        self.tbAccessLevel.setMenu(self.mAccessLevel)

        # Store actions to reopen panels
        self.panelActions = {}

        text = "Connect to server"
        self.acServerConnect = QAction(icons.remote, f"&{text}", self)
        self.acServerConnect.setStatusTip(text)
        self.acServerConnect.setToolTip(text)
        self.acServerConnect.setCheckable(True)
        self.acServerConnect.triggered.connect(self.onConnectionButtonPress)

        text = "Exit"
        self.acExit = QAction(icons.exit, f"&{text}", self)
        self.acExit.setStatusTip(text)
        self.acExit.setToolTip(text)
        self.acExit.setShortcut('Ctrl+Q')
        self.acExit.triggered.connect(self.close)

        self.acHelpAbout = QAction("About", self)
        self.acHelpAbout.triggered.connect(self.onHelpAbout)

        self.acSessionInfo = QAction("Session Information", self)
        self.acSessionInfo.triggered.connect(self.onSessionInfo)

        self.acClientTopology = QAction(
            icons.deviceInstance, "Client Topology", self)
        self.acClientTopology.triggered.connect(self.onClientTopology)

        self.acDevelopmentOverview = QAction(
            icons.yes, "Development Servers", self)
        self.acDevelopmentOverview.triggered.connect(
            self.onDevelopmentOverview)

        self.acConfig = QAction("Application Configuration", self)
        self.acConfig.triggered.connect(self.onConfiguration)

        self.acWizard = QAction("Tips'N'Tricks", self)
        self.acWizard.triggered.connect(self.onWizard)

        self.acHelpAboutQt = QAction("About Qt", self)
        self.acHelpAboutQt.triggered.connect(qApp.aboutQt)

        self.acCheckUpdates = QAction("Check for Updates", self)
        self.acCheckUpdates.triggered.connect(self.onCheckUpdates)

        self.acCheckProject = QAction("Check for Project Duplicates", self)
        self.acCheckProject.triggered.connect(self.onInvestigateProject)

        self.acDeviceTopology = QAction("Check for Project Topology", self)
        self.acDeviceTopology.triggered.connect(self.onShowDeviceTopology)

        self.acGrafana = QAction(icons.grafana, "Grafana", self)
        self.acGrafana.triggered.connect(self.onGrafana)
        self.acGrafana.setCheckable(False)

        self.acNpy2CSV = QAction(
            icons.filein, "Convert Numpy file to CSV file", self)
        self.acNpy2CSV.triggered.connect(self.onNumpyFileToCSV)

        self.acSaveConcertFile = QAction(
            icons.fileout, "Save As Karabo Concert File", self)
        self.acSaveConcertFile.triggered.connect(self.onCreateConcertFile)

        self.acSaveConcertShortcut = QAction(
            icons.saveAs, "Save As Karabo Concert Desktop Shortcut", self)
        self.acSaveConcertShortcut.triggered.connect(
            self.onCreateConcertShortcut)

        self.acOpenConcert = QAction(
            icons.load, "Open Karabo Concert File", self)
        self.acOpenConcert.triggered.connect(self.onReadConcertFile)

        self.acGuiDocumentation = QAction(
            icons.weblink, "GUI Documentation", self)
        self.acGuiDocumentation.triggered.connect(self.onGuiDocumentation)
        self.acGuiDocumentation.setCheckable(False)

        self.acKaraboDocumentation = QAction(
            icons.logo, "Karabo Documentation", self)
        self.acKaraboDocumentation.triggered.connect(
            self.onKaraboDocumentation)
        self.acKaraboDocumentation.setCheckable(False)

        self.acMacroDocumentation = QAction(
            icons.logo, "HowToMacro Documentation", self)
        self.acMacroDocumentation.triggered.connect(
            self.onMacroDocumentation)
        self.acMacroDocumentation.setCheckable(False)

        self.acShortCuts = QAction(icons.weblink, "Keyboard Shortcuts", self)
        self.acShortCuts.triggered.connect(self.showShortcuts)

        self.acRemoveUserLogin = QAction(
            icons.removeUser, "Clear User Login", self)
        self.acRemoveUserLogin.triggered.connect(self.clearUserLogin)

        self.acRegisterApplication = QAction('Register Application', self)
        self.acRegisterApplication.triggered.connect(
            self.onRegisterApplication)

    def _setupMenuBar(self):
        menuBar = self.menuBar()

        mFileMenu = menuBar.addMenu("&File")
        mFileMenu.addAction(self.acOpenConcert)
        mFileMenu.addAction(self.acSaveConcertFile)
        if const.IS_LINUX_SYSTEM:
            mFileMenu.addAction(self.acSaveConcertShortcut)

        # provide settings menu inside file

        mSettingsMenu = mFileMenu.addMenu(SETTINGS_TITLE)
        self.settingsMenus = {SETTINGS_TITLE: mSettingsMenu}

        self.acEnableHighDPI = QAction('Enable HighDPI', self)
        self.acEnableHighDPI.setCheckable(True)
        enable = get_config()["highDPI"]
        self.acEnableHighDPI.setChecked(enable)
        self.acEnableHighDPI.triggered.connect(self._store_dpi_setting)
        mSettingsMenu.addAction(self.acEnableHighDPI)

        mGeometryMenu = mSettingsMenu.addMenu(GEOMETRY_TITLE)
        self.acStoreMainWindowGeometry = QAction('Store', self)
        self.acStoreMainWindowGeometry.triggered.connect(
            self.onStoreMainWindowGeometry)
        mGeometryMenu.addAction(self.acStoreMainWindowGeometry)

        self.acEraseMainWindowGeometry = QAction('Erase', self)
        self.acEraseMainWindowGeometry.triggered.connect(
            self.onEraseMainWindowGeometry)
        mGeometryMenu.addAction(self.acEraseMainWindowGeometry)

        mFileMenu.addAction(self.acRemoveUserLogin)
        if not const.IS_MAC_SYSTEM:
            mFileMenu.addAction(self.acRegisterApplication)

        # finalize file with exit action
        mFileMenu.addSeparator()
        mFileMenu.addAction(self.acExit)

        mHelpMenu = menuBar.addMenu("&Links")
        mHelpMenu.addAction(self.acGrafana)
        mHelpMenu.addAction(self.acGuiDocumentation)
        mHelpMenu.addAction(self.acMacroDocumentation)
        mHelpMenu.addAction(self.acKaraboDocumentation)

        mViewMenu = menuBar.addMenu("&View")
        mViewMenu.addAction(self.acConfig)
        mViewMenu.addAction(self.acClientTopology)
        mViewMenu.addAction(self.acDevelopmentOverview)

        mToolsMenu = menuBar.addMenu("&Tools")
        mToolsMenu.addAction(self.acCheckProject)
        mToolsMenu.addAction(self.acNpy2CSV)

        mHelpMenu = menuBar.addMenu("&Help")
        mHelpMenu.addAction(self.acHelpAbout)
        mHelpMenu.addAction(self.acHelpAboutQt)
        mHelpMenu.addAction(self.acWizard)
        mHelpMenu.addAction(self.acCheckUpdates)
        mHelpMenu.addAction(self.acShortCuts)

        if get_config()["development"]:
            mDevMenu = menuBar.addMenu("&Developer")
            mDevMenu.addAction(self.acSessionInfo)
            mDevMenu.addAction(self.acDeviceTopology)

    def _setupToolBar(self):

        self.toolbar = toolbar = self.addToolBar("Standard")
        toolbar.setObjectName(generateObjectName(toolbar))
        toolbar.addAction(self.acExit)

        toolbar.addSeparator()
        toolbar.addAction(self.acServerConnect)

        toolbar.addWidget(self.tbAccessLevel)
        toolbar.addAction(self.tbUserSession)

        self.notification_banner = MainWindowBanner()

        # Add a widget (on the right side) for displaying network performance
        self.ui_big_data = QLabel()
        self.ui_big_data.setMaximumWidth(600)
        self.ui_big_data.setAlignment(Qt.AlignRight)

        self.update_gui_extensions = QPushButton(self)
        object_name = generateObjectName(self.update_gui_extensions)
        self.update_gui_extensions.setObjectName(object_name)
        self.update_gui_extensions.setText("Update Extensions")
        self.update_gui_extensions.setFixedHeight(40)
        self.update_gui_extensions.setStyleSheet(
            f"QPushButton#{object_name} "
            f"{{background-color:#4B8AA4; font: 13pt;}}")
        self.update_gui_extensions.setToolTip(
            "A newer version of GUI Extensions is available. Click to update")
        self.update_gui_extensions.clicked.connect(
            self._on_update_gui_extensions)

        self.ui_lamp = QLabel()
        self.ui_lamp.setFixedWidth(60)
        self.ui_lamp.setAlignment(Qt.AlignCenter)
        self.ui_lamp.setFrameStyle(QFrame.Box)

        toolbar.addWidget(self.notification_banner)
        toolbar.addWidget(self.ui_big_data)
        self.update_extensions_action = toolbar.addWidget(
            self.update_gui_extensions)
        self.update_extensions_action.setVisible(False)
        toolbar.addWidget(self.ui_lamp)

    def _setupStatusBar(self):
        self.log_widget = StatusLogWidget(self)
        status_bar = self.statusBar()
        status_bar.addPermanentWidget(self.log_widget, 1)

        # More information ...
        self.serverInfo = QLabel(self)
        self.serverInfo.setObjectName("serverInfo_label")
        self.serverInfo.setMinimumWidth(SERVER_INFO_WIDTH)
        self.serverInfo.setAlignment(Qt.AlignCenter)
        status_bar.addPermanentWidget(self.serverInfo)

        get_logger().info("Started Karabo GUI application ...")

    def _setupPanelAreas(self):
        """Build the main splitter structure of the main window
        """
        # Create a main splitter with left, middle, and right areas
        mainSplitter = QSplitter(Qt.Horizontal)
        mainSplitter.setContentsMargins(5, 5, 5, 5)
        left_area = QSplitter(Qt.Vertical, mainSplitter)
        middle_area = QSplitter(Qt.Vertical, mainSplitter)
        right_area = QSplitter(Qt.Vertical, mainSplitter)
        mainSplitter.setStretchFactor(0, 1)
        mainSplitter.setStretchFactor(1, 7)
        mainSplitter.setStretchFactor(2, 2)
        self.setCentralWidget(mainSplitter)

        # Set up the left area
        left = PanelContainer("Navigation", left_area)
        self._panel_areas[PanelAreaEnum.Left] = left

        # Set up the middle area
        middle = PanelContainer("Custom view", middle_area, handle_empty=True)
        self._panel_areas[PanelAreaEnum.Middle] = middle

        # Set up the right area
        right = PanelContainer("Configuration", right_area)
        self._panel_areas[PanelAreaEnum.Right] = right

    def _addPanelMenuAction(self, action, name):
        """Add a QAction to the view menu

        :param action: a QAction to a panel
        :param name: name of the submenu
        """
        viewMenus = self.viewMenus
        if name not in viewMenus:
            viewMenus[name] = viewMenus[name].addMenu(name)
            submenu = viewMenus[name]
        else:
            submenu = viewMenus[name]
        submenu.addAction(action)
        for submenu in self.viewMenus.values():
            submenu.setEnabled(not submenu.isEmpty())

    def _quit(self):
        """Check for project changes"""
        if self._should_save_project_before_closing():
            return False
        return True

    def _enable_toolbar(self, enable):
        self.acServerConnect.setEnabled(enable)
        self.tbAccessLevel.setEnabled(enable)

    def _open_panel(self, name):
        panel_info = _PANELS.get(name)
        if panel_info is None:
            return

        klass, area_enum = panel_info
        panel = klass()
        self._active_panels[name] = panel
        self.addPanel(panel, area_enum)

    def _unminimize_remaining_panels(self):
        """Reset the maximization of any child panels
        """
        any_minimized = False
        maximized_container = None
        for area_container in self._panel_areas.values():
            if not area_container.minimized:
                maximized_container = area_container
            else:
                any_minimized = True

        if any_minimized:
            assert maximized_container is not None
            if maximized_container.count() > 0:
                maximized_container.currentWidget().onMinimize()

    def _should_save_project_before_closing(self):
        """Asks for discard/save changes on modified project.
        """
        project = get_project_model().root_model
        if project is None:
            return False
        allowed = krb_access.access_role_allowed(AccessRole.PROJECT_EDIT)
        if get_db_conn().is_writing() or (project.modified and allowed):
            ask = ('Unsaved changes on project \"<b>{}</b>\" will be '
                   'permanently lost.<br /> Continue action?'
                   .format(project.simple_name))
            msg_box = QMessageBox(QMessageBox.Question, 'Unsaved project',
                                  ask, QMessageBox.Yes | QMessageBox.Cancel)
            msg_box.setDefaultButton(QMessageBox.Cancel)
            return msg_box.exec() != QMessageBox.Yes
        return False

    def _open_link(self, link):
        try:
            webbrowser.open_new(link)
        except webbrowser.Error:
            messagebox.show_error("No web browser available!", parent=self)

    def _begin_temporary_session(self):
        if not krb_access.is_authenticated():
            return
        UserSessionDialog(parent=self).show()

    # --------------------------------------
    # Qt slots

    @Slot()
    def _store_dpi_setting(self):
        enabled = get_config()['highDPI']
        get_config()['highDPI'] = not enabled
        self.acEnableHighDPI.setChecked(not enabled)
        text = ("Changing the high dpi settings requires a restart of the "
                "client application of the setting to become active.")
        messagebox.show_information(text)

    @Slot()
    def onRegisterApplication(self):
        register_protocol()
        get_logger().info("Registered the Karabo client application with "
                          f"version <b>{const.GUI_VERSION_LONG}</b> "
                          "for the <b>URL scheme</b> protocol.")

    @Slot()
    def onConfiguration(self):
        ApplicationConfigurationDialog(parent=self).open()

    @Slot()
    def onStoreMainWindowGeometry(self):
        get_config()['main_geometry'] = self.saveGeometry()

    @Slot()
    def onEraseMainWindowGeometry(self):
        del get_config()['main_geometry']

    @Slot()
    def onClientTopology(self):
        ClientTopologyDialog(parent=self).open()

    @Slot()
    def onDevelopmentOverview(self):
        DevelopmentTopologyDialog(parent=self).open()

    @Slot()
    def onHelpAbout(self):
        AboutDialog(parent=self).open()

    @Slot()
    def onSessionInfo(self):
        GuiSessionInfo(parent=self).show()

    @Slot()
    def onCheckUpdates(self):
        dialog = UpdateDialog(parent=self)
        dialog.open()

    @Slot()
    def onInvestigateProject(self):
        root = get_project_model().root_model
        if root is None:
            return
        if get_db_conn().is_reading():
            messagebox.show_information(
                "Please wait until project loading has finished.", parent=self)
            return

        @Slot(object)
        def handler(future):
            data = future.result()
            text = (
                "The view shows the number of instances for both servers and "
                "devices and their occurrence as duplicates.\n"
                "Duplicates can lead to configuration conflicts.")
            dialog = DataViewDialog(
                "Project Instance Conflicts", text, data, parent=self)
            dialog.show()

        background(check_instance_duplicates, root, callback=handler)

    @Slot()
    def onShowDeviceTopology(self):
        domain = get_config()["domain"]

        def show_device_topology(success, reply):
            if not success:
                reason, details = get_reason_parts(reply)
                messagebox.show_error(reason, details=details, parent=self)
                return

            dialog = ProjectTopologyDialog(parent=self)
            dialog.initialize(reply)
            dialog.open()

        call_device_slot(
            show_device_topology, KARABO_PROJECT_MANAGER, "slotGenericRequest",
            domain=domain, type="listDomainWithDevices")

    @Slot()
    def onGrafana(self):
        self._open_link(GRAFANA_LINK)

    @Slot()
    def onGuiDocumentation(self):
        self._open_link(PAGES_GUI_LINK)

    @Slot()
    def onKaraboDocumentation(self):
        self._open_link(PAGES_FW_LINK)

    @Slot()
    def onMacroDocumentation(self):
        self._open_link(PAGES_MACRO_LINK)

    @Slot()
    def onWizard(self):
        TipsTricksWizard(parent=self).open()

    @Slot(QAction)
    def onChangeAccessLevel(self, action):
        level = action.data()
        assert isinstance(level, AccessLevel), 'Garbage access level value!'
        krb_access.GLOBAL_ACCESS_LEVEL = level

        # Tell the world
        broadcast_event(KaraboEvent.AccessLevelChanged, {})

    @Slot(bool)
    def onConnectionButtonPress(self, connect):
        """Slot triggered when the `acServerConnect` button is clicked.
        i.e. The user wishes to (dis)connect the GUI client from/to the server.
        """
        if not connect and self._should_save_project_before_closing():
            # Disconnecting AND need to save first
            self.acServerConnect.setChecked(True)
        else:
            self.update_server_connection()
            # Either connecting or no need to save before disconnecting
            get_network().onServerConnection(connect, parent=self)

    @Slot(str)
    def onPanelClose(self, name):
        action = self.panelActions.get(name)
        if action is not None:
            action.setEnabled(True)

        self._active_closable_panels.pop(name)

    @Slot(float, bool)
    def onNetworkPerformance(self, proc_delay, active):
        """Color our network lamp with respect to the processing delay
        """
        color = get_processing_color(proc_delay=proc_delay)
        self.ui_lamp.setStyleSheet(f"background-color: rgba{color}")

        # The CHOOCH from the QDialog will still set the diff!
        if active:
            self.ui_lamp.setText(f'{proc_delay:.3f}')

    @Slot(bool)
    def onServerConnectionChanged(self, isConnected):
        """Slot triggered when the network connection goes up/down. At this
        point, there is no way to easily undo the state change.

        NOTE: If you need to do something before the state changes, see
        ``self.onConnectionButtonPress()``
        """
        # Un-minimize all panels when disconnecting!
        if not isConnected:
            self._unminimize_remaining_panels()
            # Erase lamp information
            self.ui_lamp.setStyleSheet("")
            self.ui_lamp.clear()
            self.ui_big_data.setText("")
            self.notification_banner.set_banner("")

        if isConnected:
            text = "Disconnect from server"
        else:
            text = "Connect to server"

        self.acServerConnect.setStatusTip(text)
        self.acServerConnect.setToolTip(text)

        self.acServerConnect.blockSignals(True)
        self.acServerConnect.setChecked(isConnected)
        self.acServerConnect.blockSignals(False)

        self.tbAccessLevel.setEnabled(isConnected)

    @Slot(bool)
    def onTemporarySession(self, _):
        if krb_access.is_temporary_session():
            ask = ACCESS_LEVEL_INFO.format(
                user=krb_access.TEMPORARY_SESSION_USER,
                access_level=krb_access.GLOBAL_ACCESS_LEVEL.name)
            dialog = QMessageBox(
                QMessageBox.Question, 'Temporary Session', ask,
                QMessageBox.Yes | QMessageBox.No, parent=self)
            move_to_cursor(dialog)
            if dialog.exec() == QMessageBox.Yes:
                get_network().endTemporarySession()
        else:
            self._begin_temporary_session()

    @Slot()
    def onNumpyFileToCSV(self):
        """Open a file dialog to select the numpy file and convert it to a csv
        """
        path = get_config()["data_dir"]
        directory = path if path and Path(path).is_dir() else ""
        file_name = getOpenFileName(
            parent=self, caption="Select Numpy file", directory=directory,
            filter="Plot Data  (*.npy *.npz)")
        if file_name:
            output_file = convert_npy_to_csv(file_name=file_name, parent=self)
            if output_file is not None:
                text = f"Converted to {output_file}"
                messagebox.show_information(text=text, parent=self)

            get_config()["data_dir"] = str(Path(file_name).parent)

    @Slot()
    def onCreateConcertFile(self):
        """Write the yaml for the concert. It should have uuid and position
        of the all opened scene panels."""
        scene_data = get_restore_data()
        if not scene_data:
            messagebox.show_information(text="No Project Scenes are open",
                                        parent=self)
            return
        directory = get_config()["data_dir"]
        file_name = getSaveFileName(
            caption="Save Concert file", filter="Yaml file(*.yaml *.yml)",
            suffix="yaml", directory=directory, parent=self)
        if file_name:
            save_concert_file(file_name, scene_data)
            get_config()["data_dir"] = str(Path(file_name).parent)
            return file_name

    @Slot()
    def onCreateConcertShortcut(self):
        file_name = self.onCreateConcertFile()
        if file_name is None:
            return
        file_name = Path(file_name)
        name = f"karabo-concert-{file_name.stem}"
        command = f'{shutil.which("conda")} ' \
                  f'run {shutil.which("karabo-concert")} {file_name}'
        icon = Path(icons.__file__).parent / "app_logo.png"
        source_path = create_linux_desktop_file(
            name=name, icon=icon, command=command, scheme="karabo")
        desktop_file = Path.home() / "Desktop" / f"{name}.desktop"
        source_path.rename(desktop_file)
        mode = 0o755  # "u=rwx,g=rx,o=rx"
        desktop_file.chmod(mode)

    @Slot()
    def onReadConcertFile(self):
        path = get_config()["data_dir"]
        directory = path if path and Path(path).is_dir() else ""
        file_name = getOpenFileName(
            caption="Run Concert File", filter="Yaml file(*.yaml *.yml)",
            directory=directory, parent=self)
        if file_name:
            run_concert(file_name=file_name)
            get_config()["data_dir"] = str(Path(file_name).parent)

    @Slot()
    def clearUserLogin(self) -> None:
        del get_config()["refresh_token"]
        del get_config()["refresh_token_user"]
        get_logger().info("Removed the user login information")

    @Slot()
    def showShortcuts(self) -> None:
        self._open_link(KEYBOARD_SHORTCUTS)

    @Slot()
    def _on_update_gui_extensions(self):
        self.update_extensions_action.setVisible(False)
        self.onCheckUpdates()
